"""Register metadata generators for a configurable writer
"""


def register_all_metadata_generators():
    """Register all metadata generators in a bliss session for
    the scan writers (currently only one).
    """
    from bliss.scanning import scan_meta  # avoid bliss patching on import
    from . import publish  # Avoid the following circular import:

    #   blisswriter.bliss.metadata
    #    -> bliss.common.auto_filter
    #       -> bliss.scanning.scan
    #          -> bliss.scanning.writer

    user_scan_meta = scan_meta.get_user_scan_meta()
    publish.register_metadata_generators(user_scan_meta)



from typing import Callable
from time import time


class PeriodicTask:
    def __init__(self, task: Callable[[], None], period: float = 0):
        self._tm0 = time()
        self._period = period
        self._task = task

    def reset(self):
        self._tm0 = time()

    def execute(self):
        tm = time()
        if (tm - self._tm0) > self._period:
            self._task()
            self._tm0 = tm



"""Save options for a writer which does not rely on extra metadata published by Bliss"""

from typing import Dict, Any
import enum


@enum.unique
class PROFILE_PARAMETERS(enum.IntEnum):
    OFF = enum.auto()
    CPU30 = enum.auto()
    CPU50 = enum.auto()
    CPU100 = enum.auto()
    WALL30 = enum.auto()
    WALL50 = enum.auto()
    WALL100 = enum.auto()
    MEM30 = enum.auto()
    MEM50 = enum.auto()
    MEM100 = enum.auto()

    @property
    def arguments(self) -> dict:
        return _PROFILE_ARGUMENTS[self]


_PROFILE_ARGUMENTS = {
    PROFILE_PARAMETERS.OFF: {},
    PROFILE_PARAMETERS.CPU30: {
        "memory": False,
        "time": True,
        "clock": "cpu",
        "timelimit": 30,
    },
    PROFILE_PARAMETERS.CPU50: {
        "memory": False,
        "time": True,
        "clock": "cpu",
        "timelimit": 50,
    },
    PROFILE_PARAMETERS.CPU100: {
        "memory": False,
        "time": True,
        "clock": "cpu",
        "timelimit": 100,
    },
    PROFILE_PARAMETERS.WALL30: {
        "memory": False,
        "time": True,
        "clock": "wall",
        "timelimit": 30,
    },
    PROFILE_PARAMETERS.WALL50: {
        "memory": False,
        "time": True,
        "clock": "wall",
        "timelimit": 50,
    },
    PROFILE_PARAMETERS.WALL100: {
        "memory": False,
        "time": True,
        "clock": "wall",
        "timelimit": 100,
    },
    PROFILE_PARAMETERS.MEM30: {"memory": True, "time": False, "memlimit": 30},
    PROFILE_PARAMETERS.MEM50: {"memory": True, "time": False, "memlimit": 50},
    PROFILE_PARAMETERS.MEM100: {"memory": True, "time": False, "memlimit": 100},
}


def _resource_profiling_from_string(s):
    try:
        return PROFILE_PARAMETERS[s.upper()]
    except KeyError as e:
        raise ValueError(str(e))


cli_saveoptions = {
    "keepshape": {
        "dest": "flat",
        "action": "store_false",
        "help": "Keep shape of multi-dimensional grid scans",
    },
    "multivalue_positioners": {
        "dest": "multivalue_positioners",
        "action": "store_true",
        "help": "Group positioners values",
    },
    "enable_external_nonhdf5": {
        "dest": "allow_external_nonhdf5",
        "action": "store_true",
        "help": "Enable external non-hdf5 files like edf (ABSOLUTE LINK!)",
    },
    "disable_external_hdf5": {
        "dest": "allow_external_hdf5",
        "action": "store_false",
        "help": "Disable external hdf5 files (virtual datasets)",
    },
    "copy_non_external": {
        "dest": "copy_non_external",
        "action": "store_true",
        "help": "Copy data instead of saving the uri when external linking is disabled",
    },
    "resource_profiling": {
        "dest": "resource_profiling",
        "default": PROFILE_PARAMETERS.OFF,
        "type": _resource_profiling_from_string,
        "choices": list(PROFILE_PARAMETERS),
        "help": "Enable resource profiling",
    },
    "required_disk_space": {
        "dest": "required_disk_space",
        "default": 200,
        "type": int,
        "help": "Required disk space in MB",
    },
    "recommended_disk_space": {
        "dest": "recommended_disk_space",
        "default": 1024,
        "type": int,
        "help": "Recommended disk space in MB",
    },
}


def extract_default_saveoptions(cli_saveoptions: Dict[str, dict]) -> Dict[str, Any]:
    saveoptions = {}
    for options in cli_saveoptions.values():
        if "default" in options:
            v = options["default"]
        else:
            v = options["action"] == "store_false"
        saveoptions[options["dest"]] = v
    return saveoptions


def default_saveoptions() -> Dict[str, Any]:
    return extract_default_saveoptions(cli_saveoptions)



"""Nexus API
"""

import os
import errno
import re
import pprint
import logging
import warnings
import types
import datetime
from functools import wraps
from io import StringIO
from collections import Counter
from contextlib import contextmanager
from typing import Mapping, Sequence

import h5py
import numpy

try:
    import gevent
    from gevent.monkey import is_anything_patched
except ImportError:

    def is_anything_patched():
        return False


from silx.io import h5py_utils
from silx.io import dictdump

from .io_utils import mkdir
from ..utils import data_merging
from ..utils.process_utils import file_processes
from ..utils.sync_utils import SharedLockPool
from ..utils.exceptions import iter_chained_exception


DEFAULT_PLOT_NAME = "plotselect"

try:
    h5py.VirtualLayout
except AttributeError:
    HASVIRTUAL = False
else:
    HASVIRTUAL = True

ISSUE_1641_FIXED = h5py.version.version_tuple >= (3, 9)  # h5py issue #1641

logger = logging.getLogger(__name__)

nxcharUnicode = h5py.special_dtype(vlen=str)
nxcharBytes = h5py.special_dtype(vlen=bytes)


class NexusInstanceExists(Exception):
    pass


def _hdf5_string_encode(s):
    """Convert a scalar or array of bytes or unicode strings to a scalar
    or array of bytes. Needed for `asNxChar`
    """
    if isinstance(s, bytes):
        return s
    elif isinstance(s, str):
        return s.encode("utf-8", errors="surrogateescape")
    elif isinstance(s, numpy.ndarray):
        if numpy.issubdtype(s.dtype, numpy.bytes_):
            return s
        elif numpy.issubdtype(s.dtype, numpy.str_) and s.ndim == 0:
            return s.encode("utf-8", errors="surrogateescape")
    return [_hdf5_string_encode(si) for si in s]


def asNxChar(s, raiseExtended=True):
    """
    Convert to Variable-length string (array or scalar).
    Uses UTF-8 encoding when possible, otherwise byte-strings
    are used (unless raiseExtended is set).

    :param s: string or sequence of strings
              string types: unicode, bytes, fixed-length numpy
    :param bool raiseExtended: raise UnicodeDecodeError for bytes
                               with extended ASCII encoding
    :returns np.ndarray(nxcharUnicode or nxcharBytes):
    :raises UnicodeDecodeError: extended ASCII encoding
    """
    if isinstance(s, numpy.ndarray) and s.dtype in (nxcharBytes, nxcharUnicode):
        return s
    try:
        # dtype=nxcharUnicode will not attempt decoding bytes
        # so readers will get UnicodeDecodeError when bytes
        # are extended ASCII encoded. So do this instead:
        numpy.array(s, dtype=str)
    except UnicodeDecodeError:
        # Reason: byte-string with extended ASCII encoding (e.g. Latin-1)
        # Solution: save as byte-string or raise exception
        # Remark: Clients will read back the data exactly as it is written.
        #         However the HDF5 character set is h5py.h5t.CSET_ASCII
        #         which is strictly speaking not correct.
        if raiseExtended:
            raise
        return numpy.array(_hdf5_string_encode(s), dtype=nxcharBytes)
    else:
        return numpy.array(s, dtype=nxcharUnicode)


def isNxChar(data):
    """String from the Nexus point of view"""
    if isinstance(data, (str, bytes)):
        return True
    elif isinstance(data, numpy.ndarray):
        return numpy.issubdtype(data.dtype, numpy.bytes_) or numpy.issubdtype(
            data.dtype, numpy.str_
        )
    else:
        return False


def asNxType(data):
    """Convert data to Nexus data type"""
    if isNxChar(data):
        return asNxChar(data)
    elif isinstance(data, (list, tuple)) and data:
        if all([isinstance(v, (str, bytes)) for v in data]):
            return asNxChar(data)
    return data


def createNxValidate(createkws):
    if "data" in createkws:
        createkws["data"] = asNxType(createkws["data"])
    dtype = createkws.get("dtype")
    if dtype is not None:
        if dtype is str:
            createkws["dtype"] = nxcharUnicode
            createkws.pop("fillvalue")
        elif dtype is bytes:
            createkws["dtype"] = nxcharBytes
            createkws.pop("fillvalue")


def ensure_timezone(tm):
    """Make sure the datetime has a timezone.
    Take the current system timezone when missing.

    :param datetime tm:
    :returns datetime:
    """
    if tm.tzinfo is None:
        return tm.astimezone()
    else:
        return tm


def datetime_to_nexus(tm):
    """
    Format datetime for Nexus saving

    :param datetime.datetime tm:
    :returns np.ndarray: variable length string
    """
    return asNxType(ensure_timezone(tm).isoformat())


def timestamp():
    """
    Timestamp for Nexus saving

    :returns np.ndarray: variable length string
    """
    return datetime_to_nexus(datetime.datetime.now())


def hdf5_sep(func):
    @wraps(func)
    def as_os_path(*args, sep="/"):
        args = [re.sub(r"[\/]+", os.sep, x) for x in args]
        ret = func(*args)
        if isinstance(ret, str):
            ret = ret.replace(os.sep, sep)
        return ret

    return as_os_path


@hdf5_sep
def hdf5_normpath(path):
    return os.path.normpath(path)


@hdf5_sep
def hdf5_relpath(path, refpath):
    return os.path.relpath(path, refpath)


@hdf5_sep
def hdf5_split(path):
    return os.path.split(path)


@hdf5_sep
def hdf5_basename(path):
    return os.path.basename(path)


@hdf5_sep
def hdf5_dirname(path):
    return os.path.dirname(path)


@hdf5_sep
def hdf5_join(*args):
    return os.path.join(*args)


def splitUri(uri):
    """
    Split Uniform Resource Identifier (URI)

    :param str uri: URI
    :return tuple: filename(str), group(str)
    """
    try:
        i = uri.index("::")
    except (IndexError, ValueError):
        filename = uri
        h5groupname = "/"
    else:
        filename = uri[:i]
        h5groupname = uri[i + 2 :]
        if h5groupname[0] != "/":
            h5groupname = "/" + h5groupname
    return filename, h5groupname


def normUri(uri):
    """Normalize uri

    :param str uri: URI
    :return str:
    """
    filename, path = splitUri(uri)
    filename = os.path.normpath(filename)
    path = hdf5_normpath(path)
    return filename + "::" + path


def _dereference_node(parent, name):
    """
    :param h5py.Dataset or h5py.Group parent:
    :param str name: appended to uri
    :returns str: uri
    """
    try:
        lnk = parent.get(name, default=None, getlink=True)
    except (KeyError, RuntimeError):
        return hdf5_join(getUri(parent), name)
    else:
        if isinstance(lnk, h5py.SoftLink):
            path = lnk.path
            if not path.startswith("/"):
                parts = name.split("/")[:-1]
                path = hdf5_join(parent.name, *parts, path)
            return parent.file.filename + "::" + hdf5_normpath(path)
        elif isinstance(lnk, h5py.ExternalLink):
            efilename = lnk.filename
            if not os.path.isabs(efilename):
                dirname = os.path.dirname(parent.file.filename)
                efilename = os.path.normpath(os.path.join(dirname, efilename))
            return efilename + "::" + lnk.path
        else:
            return hdf5_join(getUri(parent), name)


def dereference(uri, name=None, current=None):
    """Get full URI of dataset or group.
    Follows links until destination is found.

    :param h5py.Dataset or h5py.Group or str uri:
    :param str name: appended to uri
    :param File current:
    :returns str: uri
    """
    if not isinstance(uri, (str, bytes)):
        if current is None:
            current = uri.file
        uri = getUri(uri)
    if name:
        uri = hdf5_join(uri, name)
    filename, path = splitUri(uri)
    uri2 = uri
    istart = 1
    try:
        with _open_when_needed(filename, current=current, mode="r") as f:
            while True:
                parts = path.split("/")[1:]
                for i in range(istart, len(parts) + 1):
                    path2 = hdf5_join(*parts[:i])
                    filename2, path2 = splitUri(_dereference_node(f, path2))
                    path2 = hdf5_join(path2, *parts[i:])
                    uri2 = filename2 + "::" + path2
                    if uri != uri2:
                        if filename == filename2:
                            path = path2
                            istart = i + 1
                            break
                        else:
                            return dereference(uri2, current=f)
                else:
                    break
    except OSError:
        pass
    return uri2


def getUri(node):
    """
    Get full URI of dataset or group

    :param h5py.Dataset or h5py.Group:
    :returns str:
    """
    return node.file.filename + "::" + node.name


def relUri(uri, refuri):
    """
    Get uri relative to another uri

    :param str or 2-tuple uri:
    :param str or 2-tuple refuri:
    :returns (str, str):
    """
    if isinstance(uri, tuple):
        a, b = uri
    else:
        a, b = splitUri(uri)
    if isinstance(refuri, tuple):
        refa, refb = refuri
    else:
        refa, refb = splitUri(refuri)
    if a == refa:
        reta = "."
        retb = hdf5_relpath(b, refb)
    else:
        abase = os.path.basename(a)
        adir = os.path.dirname(a)
        refadir = os.path.dirname(refa)
        if adir and refadir:
            adir = os.path.relpath(adir, refadir)
        reta = os.path.join(adir, abase)
        retb = b
    return reta, retb


@contextmanager
def _open_when_needed(filename, current=None, **kwargs):
    is_current = False
    if current is not None:
        try:
            is_current = os.path.realpath(current.filename) == os.path.realpath(
                filename
            )
        except OSError:
            is_current = False
    if is_current:
        yield current
    else:
        with File(filename, **kwargs) as f:
            yield f


@contextmanager
def uriContext(uri, **kwargs):
    """
    HDF5 group/dataset context

    :param str uri:
    :yields h5py.Group/h5py.Dataset:
    """
    filename, path = splitUri(uri)
    try:
        with _open_when_needed(filename, **kwargs) as f:
            try:
                item = f[path]
            except KeyError:
                item = None
            yield item
    except OSError as e:
        logger.error("ERROR opening %s (%s)", uri, e)
        yield None


def exists(uri, current=None):
    """
    Check whether URI exists

    :param str uri:
    :param File current:
    :returns bool:
    """
    try:
        with uriContext(uri, current=current, mode="r") as node:
            return node is not None
    except (OSError, KeyError) as e:
        logger.warning(str(e))
        pass
    return False


def uriContains(uri, children=None, attributes=None, current=None):
    """
    :param str uri:
    :param list children:
    :param list attributes:
    :param File current:
    :returns bool:
    """
    if not children:
        children = []
    if not attributes:
        attributes = []
    try:
        with uriContext(uri, current=current, mode="r") as parent:
            for child in children:
                if child not in parent:
                    return False
            if attributes:
                attrs = parent.attrs
                for attr in attributes:
                    if attr not in attrs:
                        return False
    except Exception:
        return False
    return True


def nxComplete(uri):
    """
    :param str uri:
    :returns bool:
    """
    return uriContains(uri, children=["end_time"])


def iterup(h5group, includeself=True):
    """
    Iterator which yields all parent h5py.Group's up till root

    :param h5py.Group h5group:
    :param bool includeself:
    :returns generator:
    """
    if includeself:
        yield h5group
    while h5group.parent != h5group:
        h5group = h5group.parent
        yield h5group


def isLink(parent, name):
    """
    Check whether node is h5py.SoftLink or h5py.ExternalLink

    :param h5py.Group parent:
    :param str name:
    :returns bool:
    """
    try:
        lnk = parent.get(name, default=None, getlink=True)
    except (KeyError, RuntimeError):
        return False
    else:
        return isinstance(lnk, (h5py.SoftLink, h5py.ExternalLink))


def h5Name(h5group):
    """
    HDF5 Dataset of Group name

    :param h5py.Group h5group:
    :returns str:
    """
    return h5group.name.split("/")[-1]


def as_str(s):
    """
    :param str or bytes s:
    :returns str:
    """
    try:
        return s.decode()
    except AttributeError:
        return s


def attr_as_str(node, attr, default):
    """
    Get attribute value (scalar or sequence) with type `str`.

    :param h5py.Group or h5py.Dataset node:
    :param str attr:
    :param default:
    :return str or sequence(str):
    """
    v = node.attrs.get(attr, default)
    if isinstance(v, str):
        return v
    elif isinstance(v, bytes):
        return v.decode()
    elif isinstance(v, tuple):
        return tuple(as_str(s) for s in v)
    elif isinstance(v, list):
        return list(as_str(s) for s in v)
    elif isinstance(v, numpy.ndarray):
        return numpy.array(list(as_str(s) for s in v))
    else:
        return v


def nxClass(h5group):
    """
    Nexus class of existing h5py.Group (None when no Nexus instance)

    :param h5py.Group h5group:
    :returns str or None:
    """
    return attr_as_str(h5group, "NX_class", None)


def isNxClass(h5group, *classes):
    """
    Nexus class of existing h5py.Group (None when no Nexus instance)

    :param h5py.Group h5group:
    :param *classes: list(str) of Nexus classes
    :returns bool:
    """
    return nxClass(h5group) in classes


def raiseIsNxClass(h5group, *classes):
    """
    :param h5py.Group h5group:
    :param *classes: list(str) of Nexus classes
    :raises RuntimeError:
    """
    if isNxClass(h5group, *classes):
        raise RuntimeError("Nexus class should not be one of these: {}".format(classes))


def raiseIsNotNxClass(h5group, *classes):
    """
    :param h5py.Group h5group:
    :param *classes: list(str) of Nexus classes
    :raises RuntimeError:
    """
    if not isNxClass(h5group, *classes):
        raise RuntimeError("Nexus class not in {}".format(classes))


def nxClassInstantiate(parent, name, nxclass, raise_on_exists=False):
    """
    Create `parent[name]` or check its nexus class.

    :param h5py.Group parent:
    :param str or None name: `None` means the
    :param str nxclass: nxclass of the parent when `name is None`
    :param bool raise_on_exists:
    :returns bool: new instance created
    :raises RuntimeError: wrong Nexus class
    :raises NexusInstanceExists:
    """
    if name is None:
        exists = nxClass(parent) is not None
        group = parent
    else:
        try:
            group = parent.create_group(name, track_order=False)
            exists = False
        except Exception as e:
            exists = True
            group = parent[name]
            if raise_on_exists:
                logger.warning(e)
                raise NexusInstanceExists(group.name)
    if exists:
        raiseIsNotNxClass(group, nxclass)
    return not exists


def updated(h5group, final=False, parents=False):
    """
    h5py.Group has changed

    :param h5py.Group h5group:
    :param bool final: mark as final update
    :param bool parents: mark parents as well
    """
    tm = timestamp()
    for group in iterup(h5group):
        nxclass = nxClass(group)
        if nxclass is None:
            if parents:
                continue
            else:
                break
        elif nxclass in ["NXentry", "NXsubentry"]:
            if "start_time" not in group:
                group["start_time"] = tm
            if final:
                updateDataset(group, "end_time", tm)
        elif nxclass in ["NXprocess", "NXnote"]:
            updateDataset(group, "date", tm)
        # h5py issue #1641
        # elif nxclass == "NXroot":
        #    updateAttribute(group, "file_update_time", tm)
        if not parents:
            break


def updateDataset(parent, name, data):
    """
    :param h5py.Group parent:
    :param str name:
    :param data:
    """
    if name in parent:
        parent[name][()] = asNxType(data)
    else:
        parent[name] = asNxType(data)


def updateAttribute(parent, name, value):
    """
    :param h5py.Group parent:
    :param str name:
    :param value:
    """
    if value is None:
        return
    try:
        parent.attrs[name] = asNxType(value)
    except Exception as e:
        if "B-tree" in str(e):
            warnings.warn(
                f"Cannot set attribute {repr(name)}: h5py issue #1641", UserWarning
            )
        else:
            raise


def nxClassInit(
    parent, name, nxclass, parentclasses=None, raise_on_exists=False, **kwargs
):
    """
    Initialize Nexus class instance. When it does exist: create
    default attributes and datasets or raise exception.

    :param h5py.Group parent:
    :param str name:
    :param str nxclass:
    :param tuple parentclasses:
    :param bool raise_on_exists:
    :param **kwargs: see `nxInit`
    :returns h5py.Group:
    :raises RuntimeError: wrong Nexus class or parent
                          not an Nexus class instance
    :raises NexusInstanceExists:
    """
    if parentclasses:
        raiseIsNotNxClass(parent, *parentclasses)
    else:
        raiseIsNxClass(parent, None)
    if nxClassInstantiate(parent, name, nxclass, raise_on_exists=raise_on_exists):
        h5group = parent[name]
        nxdict = kwargs.get("nxdict")
        if nxdict is None:
            kwargs["nxdict"] = nxdict = dict()
        nxdict["@NX_class"] = nxclass
        nxInit(h5group, **kwargs)
        updated(h5group)
        return h5group
    else:
        return parent[name]


def nxRootInit(h5group, rootattrs=None):
    """
    Initialize NXroot instance

    :param h5py.Group h5group:
    :param dict rootattrs:
    :raises ValueError: not root
    :raises RuntimeError: wrong Nexus class
    """
    if h5group.name != "/":
        raise ValueError("Group should be the root")
    if rootattrs is None:
        rootattrs = {}
    if nxClassInstantiate(h5group, None, "NXroot"):
        updateAttribute(h5group, "NX_class", "NXroot")
        updateAttribute(h5group, "file_time", timestamp())
        updateAttribute(h5group, "file_name", h5group.file.filename)
        updateAttribute(h5group, "HDF5_Version", h5py.version.hdf5_version)
        updateAttribute(h5group, "h5py_version", h5py.version.version)
        for attr, value in rootattrs.items():
            updateAttribute(h5group, attr, value)
        updated(h5group)


def _nxdict_without_removal(nxdict: dict) -> dict:
    result = dict()
    for k, v in nxdict.items():
        if isinstance(v, Mapping):
            result[k] = _nxdict_without_removal(v)
        elif v is not None:
            result[k] = v
    return result


def nxInit(h5group, nxdict=None):
    """
    Initialize NXclass instance

    :param h5py.Group h5group:
    :param dict nxdict:
    :param str update_mode:
    """
    if nxdict:
        nxdict = _nxdict_without_removal(nxdict)
        dictdump.dicttonx(nxdict, h5group, update_mode="modify")


def nxEntryInit(
    parent, name, sub=False, start_time=None, raise_on_exists=False, **kwargs
):
    """
    Initialize NXentry and NXsubentry instance

    :param h5py.Group parent:
    :param bool sub:
    :param datetime start_time: 'now' when missing
    :param bool raise_on_exists:
    :param **kwargs: see `nxInit`
    :returns h5py.Group:
    :raises RuntimeError: wrong NX_class or parent NX_class
    :raises NexusInstanceExists:
    """
    if sub:
        nxclass = "NXsubentry"
        parents = ("NXentry",)
    else:
        nxclass = "NXentry"
        parents = ("NXroot",)
    raiseIsNotNxClass(parent, *parents)
    if nxClassInstantiate(parent, name, nxclass, raise_on_exists=raise_on_exists):
        h5group = parent[name]
        if start_time is None:
            start_time = timestamp()
        else:
            start_time = datetime_to_nexus(start_time)
        nxdict = kwargs.get("nxdict")
        if nxdict is None:
            kwargs["nxdict"] = nxdict = dict()
        nxdict["@NX_class"] = nxclass
        nxdict["start_time"] = start_time
        nxInit(h5group, **kwargs)
        updated(h5group)
        return h5group
    else:
        return parent[name]


def nxProcessConfigurationInit(
    parent, date=None, configdict=None, type="json", indent=2, raise_on_exists=False
):
    """
    Initialize NXnote instance

    :param h5py.Group parent:
    :param datetime date:
    :param dict configdict:
    :param str type: 'json' or 'ini' or None (pprint)
    :param num indent: pretty-string with indent level
    :returns h5py.Group:
    :raises RuntimeError: parent not NXprocess
    :raises NexusInstanceExists:
    """
    raiseIsNotNxClass(parent, "NXprocess")
    if configdict is not None:
        data = ""
        with StringIO() as s:
            if type == "json":
                dictdump.dicttojson(configdict, s, indent=indent)
                data = s.getvalue()
            elif type == "ini":
                dictdump.dicttoini(configdict, s)
                data = s.getvalue()
        if not data:
            data = pprint.pformat(configdict, indent=indent)
            type = "txt"
    else:
        data = None
        type = None
    return nxNote(
        parent,
        "configuration",
        data=data,
        type=type,
        date=date,
        raise_on_exists=raise_on_exists,
    )


def nxProcess(parent, name, configdict=None, raise_on_exists=False, **kwargs):
    """
    Get NXprocess instance (create or raise when missing)

    :param h5py.Group parent:
    :param str name:
    :param dict configdict:
    :param **kwargs: see `nxProcessConfigurationInit`
    :returns h5py.Group:
    :raises RuntimeError: wrong Nexus class or parent not NXentry
    :raises NexusInstanceExists:
    """
    raiseIsNotNxClass(parent, "NXentry")
    if nxClassInstantiate(parent, name, "NXprocess", raise_on_exists=raise_on_exists):
        h5group = parent[name]
        updateDataset(h5group, "program", "bliss")
        updateAttribute(h5group, "NX_class", "NXprocess")
        updated(h5group)
    else:
        h5group = parent[name]
    nxProcessConfigurationInit(
        h5group, configdict=configdict, raise_on_exists=raise_on_exists, **kwargs
    )
    nxClassInit(h5group, "results", "NXcollection")
    return h5group


def nxNote(parent, name, data=None, type=None, date=None, raise_on_exists=False):
    """
    Get NXnote instance (initialize when missing)

    :param h5py.Group parent:
    :param str name:
    :param str data:
    :param str type:
    :param datetime date:
    :param bool raise_on_exists:
    :return h5py.Group:
    :raises RuntimeError: wrong Nexus class or parent
                          not an Nexus class instance
    :raises NexusInstanceExists:
    """
    raiseIsNxClass(parent, None)
    if nxClassInstantiate(parent, name, "NXnote", raise_on_exists=raise_on_exists):
        h5group = parent[name]
        updateAttribute(h5group, "NX_class", "NXnote")
        update = True
    else:
        h5group = parent[name]
        update = False
    if data is not None:
        updateDataset(h5group, "data", data)
        update = True
    if type is not None:
        updateDataset(h5group, "type", type)
        update = True
    if date:
        updateDataset(h5group, "date", datetime_to_nexus(date))
    elif update:
        updated(h5group)
    return h5group


def _isErrno(ex, errno):
    """
    :param BaseException ex:
    :param int errno:
    :returns bool:
    """
    return f"errno = {errno}" in str(ex)


def _isLockedError(ex):
    """
    :param BaseException ex:
    :returns bool:
    """
    return _isErrno(ex, errno.EAGAIN) and "unable to lock file" in str(ex)


def isLockedError(ex):
    """
    :param BaseException ex:
    :returns bool:
    """
    for exi in iter_chained_exception(ex):
        if _isLockedError(exi):
            return True
    return False


def lockedErrorMessage(filename):
    """
    :param str filename:
    :returns str:
    """
    pattern = ".+{}$".format(os.path.basename(filename))
    procs = file_processes(pattern)
    if procs:
        msg = f"The Nexus file is locked: {repr(filename)}"
        msg += "\nThese local processes are accessing the file but not necessarily locking it:"
        for _, proc in procs:
            msg += f"\n {proc}"
        msg += "\nRemote processes are not shown."
    else:
        msg = f"The Nexus file is locked by a remote process: {repr(filename)}"
    msg += "\nTerminate the locking process or use a different acquisition file in order to continue."
    # msg += " External applications should open the file in read-only mode with file locking disabled."
    return msg + "\n"


class File(h5py_utils.File):

    _LOCKPOOL = SharedLockPool()

    def __init__(self, filename, **kwargs):
        """
        :param str filename:
        :param **kwargs: see `h5py_utils.File.__init__`
        """
        with self._protect_init(filename):
            super().__init__(filename, **kwargs)

    @contextmanager
    def _protect_init(self, filename):
        """Makes sure no other file is opened/created
        or protected sections associated to the filename
        are executed.
        """
        lockname = os.path.abspath(filename)
        with self._LOCKPOOL.acquire(None):
            with self._LOCKPOOL.acquire(lockname):
                yield

    @contextmanager
    def protect(self):
        """Protected section associated to this file."""
        lockname = os.path.abspath(self.filename)
        with self._LOCKPOOL.acquire(lockname):
            yield


if File._LOCKING_MGR is not None:  # h5py < 3.5 or libhdf5 < 1.10.7 or libhdf5 == 1.12.0
    # silx >= 1.0.0 does not allow unsafe locking behaviour (old versions require a process-wide
    # environment variable to enable/disable locking)
    # assume the Nexus writer knows what it's doing and remove the check
    File._LOCKING_MGR = h5py_utils.Hdf5FileLockingManager()
    File._LOCKING_MGR._check_locking_env = types.MethodType(
        lambda self, enable: None, File._LOCKING_MGR
    )


class nxRoot(File):
    def __init__(self, filename, mode="r", rootattrs=None, **kwargs):
        """
        :param str filename:
        :param str mode:
        :param dict rootattrs: extra attributes to the default ones
        :param **kwargs: see `h5py.File.__init__`
        """
        with self._protect_init(filename):
            if mode != "r":
                mkdir(os.path.dirname(filename))
            super().__init__(filename, mode=mode, **kwargs)
            nxRootInit(self, rootattrs)


def nxEntry(root, name, **kwargs):
    """
    Get NXentry instance (initialize when missing)

    :param h5py.Group root:
    :param str name:
    :param **kwargs: see `nxEntryInit`
    :returns h5py.Group:
    """
    return nxEntryInit(root, name, sub=False, **kwargs)


def nxSubEntry(parent, name, **kwargs):
    """
    Get NXsubentry instance (initialize when missing)

    :param h5py.Group parent:
    :param str name:
    :param **kwargs: see `nxEntryInit`
    :returns h5py.Group:
    """
    return nxEntryInit(parent, name, sub=True, **kwargs)


def nxCollection(parent, name, **kwargs):
    """
    Get NXcollection instance (initialize when missing)

    :param h5py.Group parent:
    :param str name:
    :param **kwargs: see `nxClassInit`
    :returns h5py.Group:
    """
    return nxClassInit(parent, name, "NXcollection", **kwargs)


def nxInstrument(parent, name="instrument", **kwargs):
    """
    Get NXinstrument instance (initialize when missing)

    :param h5py.Group parent:
    :param str name:
    :param **kwargs: see `nxClassInit`
    :returns h5py.Group:
    """
    return nxClassInit(
        parent, name, "NXinstrument", parentclasses=("NXentry",), **kwargs
    )


def nxDetector(parent, name, **kwargs):
    """
    Get NXdetector instance (initialize when missing)

    :param h5py.Group parent:
    :param str name:
    :param **kwargs: see `nxClassInit`
    :returns h5py.Group:
    """
    return nxClassInit(
        parent, name, "NXdetector", parentclasses=("NXinstrument",), **kwargs
    )


def nxPositioner(parent, name, **kwargs):
    """
    Get NXpositioner instance (initialize when missing)

    :param h5py.Group parent:
    :param str name:
    :param **kwargs: see `nxClassInit`
    :returns h5py.Group:
    """
    return nxClassInit(
        parent, name, "NXpositioner", parentclasses=("NXinstrument",), **kwargs
    )


def nxData(parent, name, **kwargs):
    """
    Get NXdata instance (initialize when missing)

    :param h5py.Group parent:
    :param str or None name:
    :param **kwargs: see `nxClassInit`
    :returns h5py.Group:
    """
    if name is None:
        name = DEFAULT_PLOT_NAME
    return nxClassInit(parent, name, "NXdata", **kwargs)


def nxDataGetSignals(data):
    """
    Get NXdata signals (default signal first)

    :param h5py.Group data:
    :returns list(str): signal names (default first)
    """
    signal = attr_as_str(data, "signal", None)
    auxsignals = attr_as_str(data, "auxiliary_signals", None)
    if signal is None:
        lst = []
    else:
        lst = [signal]
    if auxsignals is not None:
        lst += auxsignals.tolist()
    return lst


def nxDataSetSignals(data, signals):
    """
    Set NXdata signals (default signal first)

    :param h5py.Group data:
    :param list(str) signals:
    """
    if signals:
        updateAttribute(data, "signal", signals[0])
        if len(signals) > 1:
            updateAttribute(data, "auxiliary_signals", asNxChar(signals[1:]))
        else:
            data.attrs.pop("auxiliary_signals", None)
    else:
        data.attrs.pop("signal", None)
        data.attrs.pop("auxiliary_signals", None)
    updated(data)


def _datasetMergeInfo(
    uris,
    shape=None,
    dtype=None,
    axis=0,
    newaxis=True,
    order=None,
    fill_generator=None,
    virtual_source_args=None,
    current=None,
):
    """
    Equivalent to `numpy.stack` or `numpy.concatenate` combined
    with `numpy.reshape`. Return I/O index generator for merging.

    :param list(str) uris:
    :param tuple shape:
    :param dtype dtype:
    :param int axis:
    :param bool newaxis:
    :param str order: for reshaping
    :param fill_generator:
    :param File current:
    :returns tuple, dtype, generator: merged shape
                                      merged dtype
                                      uri I/O index generator
    """
    if not isinstance(uris, (tuple, list)):
        uris = [uris]
    if virtual_source_args:
        dtypes = {virtual_source_args["dtype"]}
        shapei = virtual_source_args["shape"]
        shapes = [shapei] * len(uris)
        ushapes = {shapei}
    else:
        shapes = []
        ushapes = []
        dtypes = set()
        for uri in uris:
            with uriContext(uri, mode="r", current=current) as dset:
                dtypes.add(dset.dtype)
                shapei = dset.shape
                shapes.append(shapei)
                shapei = list(shapei)
                try:
                    shapei[axis] = None
                except IndexError:
                    pass
                ushapes.append(tuple(shapei))
    if len(set(ushapes)) > 1:
        raise RuntimeError("Cannot concatenate datasets with shapes {}".format(ushapes))
    # Merged dtype
    if dtype is None:
        dtype = sum(numpy.array(0, dtype=dt) for dt in dtypes).dtype
    # Generator for source->destination filling
    if fill_generator is None:
        shape, fill_generator = data_merging.mergeGenerator(
            uris,
            shapes,
            shape=shape,
            order=order,
            axis=axis,
            newaxis=newaxis,
            allow_advanced_indexing=False,
        )
    elif shape is None:
        raise ValueError("Specify 'shape' when specifying 'fill_generator'")
    return shape, dtype, fill_generator


def createVirtualDataset(
    h5group,
    name,
    uris,
    axis=0,
    newaxis=True,
    maxshape=None,
    fillvalue=None,
    shape=None,
    dtype=None,
    order=None,
    fill_generator=None,
    virtual_source_args=None,
):
    """
    Create a virtual dataset (references to the individual datasets)

    :param h5py.Group h5group: parent
    :param str name:
    :param list(str) uris:
    :param int axis:
    :param bool newaxis:
    :param tuple maxshape:
    :param fillvalue:
    :param dtype:
    :param str order: for reshaping
    :param fill_generator:
    :param dict virtual_source_args: arguments for VirtualSource (avoid opening the source files)
    :returns h5py.Dataset:
    """
    current = h5group.file
    shape, dtype, fill_generator = _datasetMergeInfo(
        uris,
        shape=shape,
        dtype=dtype,
        axis=axis,
        order=order,
        newaxis=newaxis,
        fill_generator=fill_generator,
        virtual_source_args=virtual_source_args,
        current=current,
    )
    layout = h5py.VirtualLayout(shape, dtype=dtype, maxshape=maxshape)
    destination = splitUri(getUri(h5group))
    yield_to_eventloop = is_anything_patched()
    if virtual_source_args:
        logger.debug("Create VDS {} without opening sources".format(repr(destination)))
        for uri, idx_generator in fill_generator():
            source = splitUri(uri)
            spath, sname = relUri(source, destination)
            # TODO: VirtualSource does not support relative
            #       file paths upwards
            if ".." in spath:
                spath, sname = source
            # TODO: VirtualSource does not support relative
            #       dataset paths like SoftLink
            sname = source[1]
            vsource = h5py.VirtualSource(spath, sname, **virtual_source_args)
            for idxin, idxout in idx_generator():
                if idxin:
                    layout[idxout] = vsource[idxin]
                else:
                    layout[idxout] = vsource
                if yield_to_eventloop:
                    gevent.sleep()
    else:
        logger.debug("Create VDS {} with opening sources".format(repr(destination)))
        for uri, idx_generator in fill_generator():
            source = splitUri(uri)
            spath, sname = relUri(source, destination)
            with uriContext(uri, mode="r", current=current) as dset:
                # TODO: VirtualSource does not support relative
                #       file paths upwards
                if ".." in spath:
                    spath, sname = source
                # TODO: VirtualSource does not support relative
                #       dataset paths like SoftLink
                sname = source[1]
                vsource = h5py.VirtualSource(
                    spath,
                    sname,
                    shape=dset.shape,
                    dtype=dset.dtype,
                    maxshape=dset.maxshape,
                )
                for idxin, idxout in idx_generator():
                    if idxin:
                        layout[idxout] = vsource[idxin]
                    else:
                        layout[idxout] = vsource
                    if yield_to_eventloop:
                        gevent.sleep()
    return h5group.create_virtual_dataset(name, layout, fillvalue=fillvalue)


def createConcatenatedDataset(
    h5group,
    name,
    uris,
    axis=0,
    newaxis=True,
    shape=None,
    dtype=None,
    order=None,
    fill_generator=None,
    **kwargs,
):
    """
    Create a concatenated dataset (copy of the individual datasets)

    :param h5py.Group h5group: parent
    :param str name:
    :param list(str) uris:
    :param int axis:
    :param bool newaxis:
    :param dtype:
    :param str order: for reshaping
    :param fill_generator:
    :returns h5py.Dataset:
    """
    current = h5group.file
    shape, dtype, fill_generator = _datasetMergeInfo(
        uris,
        shape=shape,
        dtype=dtype,
        order=order,
        axis=axis,
        newaxis=newaxis,
        fill_generator=fill_generator,
        current=current,
    )
    kwargs["shape"] = shape
    kwargs["dtype"] = dtype
    dset = h5group.create_dataset(name, **kwargs)
    yield_to_eventloop = is_anything_patched()
    for uri, index_generator in fill_generator():
        with uriContext(uri, mode="r", current=current) as dseti:
            for idxin, idxout in index_generator():
                dset[idxout] = dseti[idxin]
                if yield_to_eventloop:
                    gevent.sleep()
    return dset


def createMergedDataset(h5group, name, uris, virtual=True, **kwargs):
    """
    Merge datasets into one dataset

    :param h5py.Group h5group: parent
    :param str name:
    :param list(str) uris:
    :param bool virtual: merge as virtual dataset (copy otherwise)
    :param **kwargs:
    :returns h5py.Dataset:
    """
    if not HASVIRTUAL:
        logger.warning(
            "Virtual HDF5 datasets are not supported: concatenate instead (this creates a copy)."
        )
        virtual = False
    if virtual:
        return createVirtualDataset(h5group, name, uris, **kwargs)
    else:
        return createConcatenatedDataset(h5group, name, uris, **kwargs)


def vdsmapUri(dset, vdsmap):
    """
    :param Dataset dset:
    :param VDSmap vdsmap:
    :returns str:
    """
    filename = vdsmap.file_name
    if not os.path.isabs(filename):
        path = os.path.dirname(dset.file.filename)
        if filename == ".":
            filename = os.path.basename(dset.file.filename)
        filename = os.path.join(path, filename)
    return filename + "::" + vdsmap.dset_name


def vdsmapIsValid(dset, vdsmap):
    """
    :param Dataset dset:
    :param VDSmap vdsmap:
    :returns bool:
    """
    return exists(vdsmapUri(dset, vdsmap), current=dset.file)


def vdsIsValid(dset):
    """
    :param Dataset dset:
    :returns bool:
    """
    if not dset.is_virtual:
        raise ValueError("Not a virtual dataset")
    sources = dset.virtual_sources()
    if not sources:
        shape = dset.shape
        return not shape or not all(shape)
    for vdsmap in sources:
        if not vdsmapIsValid(dset, vdsmap):
            return False
    return True


def createLink(h5group, name, destination, abspath=False):
    """
    Create hdf5 soft (supports relative down paths)
    or external link (supports relative paths).

    :param h5py.Group h5group: location of the link
    :param str name:
    :param str or h5py.Dataset destination:
    :param bool abspath: absolute file path
    :returns: h5py link object
    """
    if not isinstance(destination, (str, bytes)):
        destination = getUri(destination)
    if "::" in destination:
        destination = splitUri(destination)
    else:
        destination = h5group.file.filename, destination
    filename, path = destination
    lnk_filename, lnk_path = relUri(destination, getUri(h5group))
    if lnk_filename == ".":
        # TODO: h5py.SoftLink does not support relative links upwards
        if ".." in lnk_path:
            lnk_path = path
        lnk = h5py.SoftLink(lnk_path)
    else:
        if abspath:
            lnk_filename = filename
        lnk = h5py.ExternalLink(lnk_filename, lnk_path)
    h5group[name] = lnk
    return lnk


def nxCreateDataSet(h5group, name, value, attrs, stringasuri=False):
    """
    Create a dataset or a link to a dataset

    :param h5py.Group h5group: parent
    :param str name: dataset name
    :param value: None: do not create dataset
                  str: create link to uri when `stringasuri==True`
                  h5py.Dataset: create link
                  dict: `isinstance(dict['data'], h5py.Dataset)`: soft or external link
                        `'axis' in dict`: merge uris in `dict['data']`
                        else: h5py.create_dataset arguments
                  other (numpy.ndarray, num, ...): data content
    :param dict attrs: set as dataset attributes
    :param str stringasuri: interpret string as a dataset uri
    :returns h5py.Dataset:
    """
    if stringasuri:
        dsettypes = (h5py.Dataset, str, bytes)
    else:
        dsettypes = (h5py.Dataset,)
    if isinstance(value, dict):
        data = value.get("data", None)
        merge = any(
            k in value for k in ["axis", "newaxis", "virtual", "fill_generator"]
        )
        if not merge:
            if isinstance(data, dsettypes):
                # dataset or uri
                shape = value.get("shape", None)
                # TODO: dtype, maxshape as well?
                # merge when attributes do not match the existing dataset
                if isinstance(data, h5py.Dataset):
                    merge = shape != data.shape
                else:
                    with uriContext(data, mode="r", current=h5group.file) as dset:
                        if dset is not None:
                            merge = shape != dset.shape
                # link when attributes match the existing dataset
                if not merge:
                    value = data
            else:
                value["data"] = asNxType(data)
    if value is None:
        # dataset exists already or will be created elsewhere
        pass
    elif isinstance(value, dsettypes):
        # link to dataset
        logger.debug(
            "Create HDF5 dataset {}/{}: link to {}".format(getUri(h5group), name, value)
        )
        createLink(h5group, name, value)
    elif isinstance(value, dict):
        # create dataset (internal, external, virtual) with extra options
        if merge:
            value = dict(value)
            axis = value.pop("axis", 0)
            uris = value.pop("data", [])
            logger.debug("Create HDF5 dataset {}/{}".format(getUri(h5group), name))
            createMergedDataset(h5group, name, uris, axis=axis, **value)
        else:
            # TODO: external datasets do not support relative paths
            # external = value.get('external', None)
            # if external:
            #    dirname = os.path.dirname(h5group.file.filename)
            #    value['external'] = [(os.path.relpath(tpl[0], dirname),) + tpl[1:]
            #                         for tpl in external]
            logger.debug("Create HDF5 dataset {}/{}".format(getUri(h5group), name))
            createNxValidate(value)
            h5group.create_dataset(name, **value)
    else:
        # create dataset (internal) without extra options
        logger.debug("Create HDF5 dataset {}/{}".format(getUri(h5group), name))
        h5group[name] = asNxType(value)
    dset = h5group.get(name, None)
    if attrs and dset is not None:
        for k, v in attrs.items():
            updateAttribute(dset, k, v)
    return dset


def nxDatasetInterpretation(scan_ndim, detector_ndim, dataset_ndim):
    """
    Dataset interpretation

    :param int scan_ndim: scan dimensions
    :param int detector_ndim: detector data
    :param int dataset_ndim: as saved (scan_ndim may be flattened)
    """
    if detector_ndim == 0:
        if scan_ndim == dataset_ndim:
            # Scan dimension is not flattened
            if scan_ndim == 2:
                return "image"
            elif scan_ndim == 3:
                return "vertex"
    elif detector_ndim == 1:
        return "spectrum"
    elif detector_ndim == 2:
        return "image"
    elif detector_ndim == 3:
        return "vertex"
    return None


def nxDataAddSignals(data, signals, append=True):
    """
    Add signals to NXdata instance

    :param h5py.Group data:
    :param list(3-tuple) signals: see `nxCreateDataSet`
    :param bool append:
    """
    raiseIsNotNxClass(data, "NXdata")
    if append:
        names = nxDataGetSignals(data)
    else:
        names = []
    for name, value, attrs in signals:
        nxCreateDataSet(data, name, value, attrs, stringasuri=True)
        if name not in names:
            names.append(name)
    if names:
        nxDataSetSignals(data, names)


def nxDataAddAxes(data, axes, append=True):
    """
    Add axes to NXdata instance

    :param h5py.Group data: NXdata instance
    :param list(3-tuple) axes: see `nxCreateDataSet`
    :param bool append:
    """
    raiseIsNotNxClass(data, "NXdata")
    if append:
        names = attr_as_str(data, "axes", [])
    else:
        names = []
    for name, value, attrs in axes:
        nxCreateDataSet(data, name, value, attrs, stringasuri=True)
        if name not in names:
            names.append(name)
    if names:
        updateAttribute(data, "axes", asNxChar(names))
        updated(data)


def nxDataAddErrors(data, errors):
    """
    For each dataset in "data", link to the corresponding dataset in "errors".

    :param h5py.Group data:
    :param h5py.Group errors:
    """
    for name in data:
        dest = errors.get(name, None)
        if dest:
            data[name + "_errors"] = h5py.SoftLink(dest.name)


def selectDatasets(root, match=None):
    """
    Select datasets with given restrictions. In case of `root`
    is an NXdata instance, an additional restriction is imposed:
    the dataset must be specified as a signal (including auxilary signals).

    :param h5py.Group or h5py.Dataset root:
    :param match: restrict selection (callable, 'max_ndim', 'mostcommon_ndim')
    :returns list(h5py.Dataset):
    """
    if match == "max_ndim":
        match, post = None, match
    elif match == "mostcommon_ndim":
        match, post = None, match
    else:
        post = None
    if not match:

        def match(dset):
            return True

    datasets = []
    if isinstance(root, h5py.Dataset):
        if match(root):
            datasets = [root]
    else:
        labels = nxDataGetSignals(root)
        if not labels:
            labels = root.keys()
        for label in labels:
            dset = root.get(label, None)
            if not isinstance(dset, h5py.Dataset):
                continue
            if match(dset):
                datasets.append(dset)
        if post == "max_ndim":
            ndimref = max(dset.ndim for dset in datasets)
        elif post == "mostcommon_ndim":
            occurences = Counter(dset.ndim for dset in datasets)
            ndimref = occurences.most_common(1)[0][0]
        else:
            ndimref = None
        if ndimref is not None:
            datasets = [dset.ndim == ndimref for dset in datasets]
    return datasets


def markDefault(h5node, nxentrylink=True):
    """
    Mark HDF5 Dataset or Group as default (parents get notified as well)

    :param h5py.Group or h5py.Dataset h5node:
    :param bool nxentrylink: Use a direct link for the default of an NXentry instance
    """
    nxdata = None
    nxclass = nxClass(h5node)
    for parent in iterup(h5node, includeself=False):
        parentnxclass = nxClass(parent)
        if parentnxclass == "NXdata":
            signals = nxDataGetSignals(parent)
            signal = h5Name(h5node)
            if signal in signals:
                signals.pop(signals.index(signal))
            nxDataSetSignals(parent, [signal] + signals)
            updated(parent)
        elif nxclass in ["NXentry", "NXsubentry"]:
            updateAttribute(parent, "default", h5Name(h5node))
            updated(parent)
        elif nxclass == "NXdata":
            updateAttribute(parent, "default", h5Name(h5node))
            updated(parent)
            nxdata = h5node
        else:
            updateAttribute(parent, "default", h5Name(h5node))
            updated(parent)
        h5node = parent
        nxclass = parentnxclass
    if nxdata is not None:
        _nxentry_plot_link(nxdata)


def _nxentry_plot_link(nxdata):
    for nxentry in iterup(nxdata, includeself=False):
        if isNxClass(nxentry, "NXentry"):
            break
    else:
        return
    plotname = DEFAULT_PLOT_NAME
    if isLink(nxentry, plotname):
        del nxentry[plotname]
    if plotname in nxentry:
        if nxentry[plotname].name == nxdata.name:
            return
        fmt = plotname + "{}"
        i = 0
        while fmt.format(i) in nxentry:
            i += 1
        plotname = fmt.format(i)
    nxentry[plotname] = h5py.SoftLink(nxdata.name)
    updateAttribute(nxentry, "default", plotname)


def getDefault(node, signal=True):
    """
    :param h5py.Group or h5py.Dataset node:
    :returns str: path of dataset or NXdata group
    """
    default = attr_as_str(node, "default", "")
    root = node.file
    if default and not default.startswith("/"):
        if node.name == "/":
            default = "/" + default
        else:
            default = node.name + "/" + default
    while default:
        try:
            node = root[default]
        except KeyError:
            break
        nxclass = attr_as_str(node, "NX_class", "")
        if nxclass == "NXdata":
            if signal:
                name = attr_as_str(node, "signal", "data")
                try:
                    default = node[name].name
                except KeyError:
                    pass
            else:
                default = node.name
            break
        else:
            add = attr_as_str(node, "default", "")
            if add.startswith("/"):
                default = add
            elif add:
                default += "/" + add
            else:
                break
    return default


def getDefaultUri(filename, signal=True, **kwargs):
    """
    :param str filename:
    :returns str: full uri to default data
    """
    with File(filename, **kwargs) as f:
        path = getDefault(f, signal=signal)
    if path:
        return filename + "::" + path
    else:
        return None


def dicttonx(treedict, *args, **kw) -> None:
    dictdump.dicttonx(data_for_hdf5(treedict), *args, **kw)


def data_for_hdf5(data):
    if isinstance(data, Mapping):
        return {key: data_for_hdf5(value) for key, value in data.items()}
    if isinstance(data, (str, bytes, numpy.ndarray)):
        return data
    if (
        isinstance(data, Sequence)
        and len(data)
        and all(isinstance(item, Mapping) for item in data)
    ):
        return _as_structured_array(data)
    return data


def _as_structured_array(data: Sequence[Mapping]) -> numpy.ndarray:
    dtypes = dict()
    for adict in data:
        for key, value in adict.items():
            if key in dtypes:
                dtypes[key] = dtypes[key] + value
            else:
                dtypes[key] = value
    defaults = {key: value * 0 for key, value in dtypes.items()}
    dtype = [
        (key, str(numpy.array(value).dtype).replace("U", "S"))
        for key, value in dtypes.items()
    ]
    values = [
        tuple(adict.get(key, default) for key, default in defaults.items())
        for adict in data
    ]
    return numpy.array(values, dtype=dtype)



from blisswriter.nexus import devices


def test_devices_shortnamemap():
    assert devices.shortnamemap([]) == {}
    assert devices.shortnamemap(["a:b:c"]) == {"a:b:c": "c"}
    assert devices.shortnamemap(["a:b:c", "a:b:d"]) == {"a:b:c": "c", "a:b:d": "d"}
    assert devices.shortnamemap(["a:b:c", "a:b:c", "a:b:d"]) == {
        "a:b:c": "c",
        "a:b:d": "d",
    }
    assert devices.shortnamemap(["a:b:c", "a:b:d", "c"]) == {
        "a:b:c": "b:c",
        "a:b:d": "d",
        "c": "c",
    }
    assert devices.shortnamemap(["a:b:c", "b:c:d", "b:c"]) == {
        "a:b:c": "a:b:c",
        "b:c:d": "d",
        "b:c": "b:c",
    }
    assert devices.shortnamemap(["a:b:c", "a:c"]) == {"a:b:c": "b:c", "a:c": "a:c"}
    assert devices.shortnamemap(["a:b:c", "b:c"]) == {"a:b:c": "a:b:c", "b:c": "b:c"}
    assert devices.shortnamemap(["b:a", "c:a"]) == {"b:a": "b:a", "c:a": "c:a"}


def test_parse_devices__lima_roi_with_underscore():
    fullname = "lima_simulator:roi_counters:roi__1_min"
    d = {
        fullname: {
            "device_type": "lima",
            "single": None,
        },
    }
    devices.parse_devices(d, True, False)
    assert d[fullname]["data_name"] == "min"
    assert d[fullname]["data_type"] == "min"
    assert d[fullname]["metadata_keys"] == {"roi__1": "selection"}



"""Ravel/unravel index
"""

import numpy


class Order:
    def __init__(self, order=None, caxes=None):
        """
        :param str order: "C": fast axis last, "F": fast axis first
        :param tuple or str caxes: No jumps along these dimensions when
                                   looping through the array in order.
                                   Supported strings: ["all"]
        """
        self.order = order
        self.caxes = caxes

    def __eq__(self, other):
        return self.order == other.order and self.caxes == other.caxes

    def __str__(self):
        if self.caxes:
            return "{} ({})".format(self.order, self.caxes)
        else:
            return self.order

    def __copy__(self):
        return self.__class__(self.order, self.caxes)

    def copy(self):
        return self.__copy__()

    @property
    def order(self):
        return self._order

    @order.setter
    def order(self, value):
        if value:
            value = value.upper()
        else:
            value = "C"
        if value not in ["C", "F"]:
            raise ValueError("Order must be C or F")
        self._order = value

    @property
    def corder(self):
        return self.order == "C"

    @property
    def forder(self):
        return self.order == "F"

    @property
    def caxes(self):
        return self.__caxes

    @caxes.setter
    def caxes(self, value):
        if isinstance(value, str):
            if value != "all":
                raise ValueError('caxes should be "all" or a tuple of integers')
            self.__caxes = value
        else:
            if value:
                self.__caxes = set(value)
            else:
                self.__caxes = set()

    def ravel(self, idx, shape):
        """
        :param tuple(array) idx: nD indices (nD x n)
        :param tuple shape:
        :returns ndarray: 1D indices (length n)
        """
        idx = self._mindex_discontinuous(idx, shape)
        return numpy.ravel_multi_index(idx, shape, order=self.order)

    def unravel(self, idx, shape):
        """
        :param array idx: 1D indices (length n)
        :param tuple shape:
        :returns tuple(ndarray): nD indices (nD x n)
        """
        idx = numpy.unravel_index(idx, shape, order=self.order)
        idx = self._mindex_continuous(idx, shape)
        return idx

    def reshape(self, arr, shape):
        """
        :param ndarray arr: nD array
        :param tuple shape: mD
        :returns ndarray: mD array
        """
        if self.caxes:
            ret = numpy.empty_like(arr, shape=shape)
            midx_get = self.unravel(range(arr.size), arr.shape)
            itget = zip(*midx_get)
            midx_set = self.unravel(range(arr.size), shape)
            itset = zip(*midx_set)
            for iget, iset in zip(itget, itset):
                ret[iset] = arr[iget]
            return ret
        else:
            return arr.reshape(shape, order=self.order)

    def flatten(self, arr):
        """
        :param ndarray arr: nD array
        :returns ndarray: 1D array
        """
        if self._caxes(arr.shape):
            midx = self.unravel(range(arr.size), arr.shape)
            return numpy.asarray([arr[idx] for idx in zip(*midx)])
        else:
            return arr.flatten(order=self.order)

    def _ravel(self, idx, shape):
        """
        Same as numpy.ravel_multi_index

        :param tuple(array) idx: nD indices (nD x n)
        :param tuple shape:
        :returns ndarray: 1D indices (length n)
        """
        # TODO: can we merge this with _mindex_discontinuous?
        mshape = self._mshape(shape)
        return sum(x * m for x, m in zip(idx, mshape))

    def _unravel(self, idx, shape):
        """
        Same as numpy.unravel_index

        :param array idx: 1D indices (length n)
        :param tuple shape:
        :returns tuple(ndarray): nD indices (nD x n)
        """
        # TODO: can we merge this with _mindex_continuous?
        mshape = self._mshape(shape)
        idx = numpy.asarray(idx)
        return tuple(idx // m % s for s, m in zip(shape, mshape))

    def _mshape(self, shape):
        if self.order == "C":
            return numpy.cumprod((1,) + shape[: -len(shape) : -1])[::-1]
        else:
            return numpy.cumprod((1,) + shape[:-1])

    def _mindex_continuous(self, idx, shape):
        """
        Convert discontinuous index to continuous

        :param tuple(array) idx: nD indices (nD x n)
        :param tuple shape:
        :param tuple(array): nD indices (nD x n)
        """
        return self._mindex_flip(idx, shape, reverse=False)

    def _mindex_discontinuous(self, idx, shape):
        """
        Convert continuous index to discontinuous

        :param tuple(array) idx: nD indices (nD x n)
        :param tuple shape:
        :param tuple(array): nD indices (nD x n)
        """
        return self._mindex_flip(idx, shape, reverse=True)

    def _mindex_flip(self, idx, shape, reverse=False):
        """
        Convert discontinuous index to continuous

        :param tuple(array) idx: nD indices (nD x n)
        :param tuple shape:
        :param tuple(array): nD indices (nD x n)
        """
        # Do we have continuous axes?
        caxes = self._caxes(shape)
        if not caxes:
            return idx
        ndim = len(shape)
        idx = tuple(idx[i].copy() for i in range(ndim))
        # Flip continuous axes
        if (self.order == "F") ^ reverse:
            it = range(ndim)
        else:
            it = range(ndim - 1, -1, -1)
        for i in it:
            if i not in caxes:
                continue
            # Count iterations based in indices in outer loops
            niteri = self._iter_index(idx, shape, i)
            # Reverse idx[i] when in an odd iteration
            arr = idx[i]
            mask = (niteri % 2) == 1
            arr[mask] = shape[i] - 1 - arr[mask]
        return idx

    def _caxes(self, shape):
        """
        Continuous axes. The last axis is iterated through only
        once so it is irrelavent.

        :param tuple shape:
        :returns tuple:
        """
        if not self.caxes:
            return tuple()
        ndim = len(shape)
        if self.order == "F":
            ilast = ndim - 1
        else:
            ilast = 0
        if self.caxes == "all":
            caxes = range(ndim)
        else:
            caxes = self.caxes
        return tuple(i for i in caxes if i != ilast)

    def _iter_index(self, idx, shape, dim):
        """
        The iteration index of dimension `dim` in which `idx` occurs.

        :param tuple(array) idx: nD indices (nD x n)
        :param tuple shape:
        :param int dim:
        :param array: iteration indices (n)
        """
        ndim = len(shape)
        if self.order == "F":
            diminc = 1
            dimlast = ndim - 1
        else:
            diminc = -1
            dimlast = 0
        # The last dimension is only iterated through
        # once so iteration index is zero
        if dim == dimlast:
            return numpy.zeros_like(idx[0])
        # First outer loop
        niter = idx[dim + diminc].copy()
        m = 1
        # Subsequent outer loops (if any)
        for j in range(dim + 2 * diminc, dimlast + diminc, diminc):
            m *= shape[j - diminc]
            niter += idx[j] * m
        return niter

    def swap_index(self, idx, shape, neworder):
        """
        Convert flat index in self order to neworder

        :param array idx: 1D indices (length n)
        :param tuple shape:
        :param Order neworder:
        :returns array: 1D indices (length n)
        """
        if self == neworder or len(shape) <= 1:
            return idx
        else:
            idx = self.unravel(idx, shape)
            return neworder.ravel(idx, shape)

    def swap_full_index(self, shape, neworder):
        """
        :param tuple shape:
        :param Order neworder:
        :returns array: 1D indices (length n)
        """
        idx = numpy.arange(numpy.prod(shape))
        if self == neworder or len(shape) <= 1:
            return idx
        elif not self._caxes(shape) and not neworder._caxes(shape):
            idx = idx.reshape(shape, order=neworder.order)
            return idx.flatten(order=self.order)
        else:
            idx = self.unravel(idx, shape)
            return neworder.ravel(idx, shape)

    def swap_mindex(self, idx, shape, neworder):
        """
        Convert unravelled index in self order to neworder

        :param tuple(array) idx: nD indices (nD x n)
        :param tuple shape:
        :param Order neworder:
        :returns array: 1D indices (length n)
        """
        if self == neworder or len(shape) <= 1:
            return idx
        else:
            idx = self.ravel(idx, shape)
            return neworder.unravel(idx, shape)

    def swap_list(self, lst, shape, neworder):
        """
        :param list lst: 1D items (length == n)
        :param tuple shape:
        :param Order neworder:
        :returns array: lst in new order
        """
        if self == neworder or len(shape) <= 1:
            return lst
        else:
            idx = neworder.swap_full_index(shape, self)
            return [lst[i] for i in idx]



import os
import abc
import logging
from contextlib import contextmanager
from typing import Tuple

import numpy

from ..utils.logging_utils import CustomLogger
from ..io import nexus


logger = logging.getLogger(__name__)


class Hdf5ItemWriter(abc.ABC):
    """Manages creation and growth of one HDF5 dataset or group."""

    def __init__(
        self, filename=None, parent=None, filecontext=None, parent_logger=None
    ):
        """
        :param str filename: HDF5 file name
        :param str filecontext: HDF5 open context manager
        :param str parent: path in the HDF5 file
        :param parent_logger:
        """
        if filecontext is None:
            filecontext = self._filecontext
        self.filename = filename
        self.filecontext = filecontext
        if parent is None:
            self.parent = ""
        else:
            self.parent = parent
        if parent_logger is None:
            parent_logger = logger
        self.logger = CustomLogger(parent_logger, self)
        self.npoints = 0
        self._exists = False
        self._parent_exists = False
        self.__cached_dataset_nxroot = None
        self.__cached_dataset = None
        self.__cached_parent_nxroot = None
        self.__cached_parent = None
        self.__cached_nxroot = None

    def __repr__(self):
        if self.name:
            return self.path
        else:
            return os.path.splitext(os.path.basename(self.filename))[0]

    @property
    def path(self):
        """Path inside the file of the dataset"""
        if self.name:
            return "/".join([self.parent, self.name])
        else:
            return self.parent

    @property
    def url(self):
        """Full URL of the dataset"""
        return self.filename + "::" + self.path

    @property
    def parent_url(self):
        """Full URL of the parent"""
        return self.filename + "::" + self.parent

    @abc.abstractproperty
    def name(self):
        """Name of the dataset inside the file"""
        pass

    @abc.abstractproperty
    def alias(self):
        """Alias of the dataset inside the file"""
        pass

    @contextmanager
    def _filecontext(self):
        if self.__cached_nxroot is None:
            with nexus.nxRoot(self.filename, mode="a") as nxroot:
                self.__cached_nxroot = nxroot
                try:
                    yield nxroot
                finally:
                    self.__cached_nxroot = None
        else:
            yield self.__cached_nxroot

    def create(self):
        """Skipped when it already exists. It may not exist after creation."""
        self.create_parent()
        if self.exists:
            return
        with self.filecontext() as nxroot:
            self._create(nxroot)

    def create_parent(self):
        """Skipped when it already exists. It may not exist after creation."""
        if self.parent_exists:
            return
        with self.filecontext() as nxroot:
            self._create_parent(nxroot)

    @abc.abstractmethod
    def _create(self, nxroot):
        pass

    @abc.abstractmethod
    def _create_parent(self, nxroot):
        pass

    @property
    def exists(self):
        """
        :returns bool:
        """
        if self._exists:
            return True
        with self.filecontext() as nxroot:
            self._exists = exists = self.path in nxroot
            return exists

    @property
    def parent_exists(self):
        """
        :returns bool:
        """
        if self._parent_exists:
            return True
        with self.filecontext() as nxroot:
            self._parent_exists = exists = self.parent in nxroot
            return exists

    @contextmanager
    def open(self, create=False):
        """
        :param bool create: when missing
        :yields h5py.Dataset or None:
        """
        with self.filecontext() as nxroot:
            cached_dataset_nxroot = nxroot
            if cached_dataset_nxroot is self.__cached_dataset_nxroot:
                yield self.__cached_dataset
            else:
                if create and not self.exists:
                    self.create()
                try:
                    dset = nxroot[self.path]
                except Exception:
                    self.logger.warning("'%s' does not exist", self.url)
                    dset = None
                else:
                    self.__cached_dataset = dset
                    self.__cached_dataset_nxroot = cached_dataset_nxroot
                yield dset

    @contextmanager
    def open_parent(self, create=False):
        """
        :param bool create: when missing
        :yields h5py.Group or None:
        """
        with self.filecontext() as nxroot:
            cached_parent_nxroot = nxroot
            if cached_parent_nxroot is self.__cached_parent_nxroot:
                yield self.__cached_parent
            else:
                if create and not self.parent_exists:
                    self.parent_create()
                try:
                    group = nxroot[self.parent]
                except Exception:
                    self.logger.warning("'%s' does not exist", self.parent_url)
                    group = None
                else:
                    self.__cached_parent = group
                    self.__cached_parent_nxroot = cached_parent_nxroot
                yield group

    def add(self, newdata):
        """Add data

        :param sequence newdata:
        """
        with self.open(create=True) as destination:
            try:
                self.npoints += self._insert_data(destination, newdata)
            except Exception:
                self.logger.error("Exception when adding data")
                raise

    @abc.abstractmethod
    def _insert_data(self, destination, newdata):
        """Insert new data in dataset

        :param h5py.Dataset or h5py.Group dset:
        :param sequence newdata:
        :returns int: number of added points
        """
        pass

    @property
    def npoints_expected(self):
        return 0

    @property
    def complete(self):
        """Variable length scans are marked complete when we have some data"""
        n, nall = self.npoints, self.npoints_expected
        return n and n >= nall

    @property
    def progress(self):
        if self.npoints_expected:
            return self.npoints / self.npoints_expected
        else:
            if self.npoints:
                return numpy.nan
            else:
                return 0

    @property
    def progress_string(self) -> Tuple[str, str]:
        if self.npoints_expected:
            sortkey = self.npoints / self.npoints_expected
            s = f"{sortkey * 100:.0f}%"
        else:
            sortkey = self.npoints
            s = f"{sortkey:d}pts"
        return s, sortkey

    @property
    def _progress_log_suffix(self):
        return ""

    def log_progress(self, expect_complete: bool = False) -> bool:
        npoints_expected = self.npoints_expected
        npoints_current = self.npoints
        complete = self.complete
        if expect_complete:
            if complete:
                self.logger.debug(
                    "%d/%d points published%s",
                    npoints_current,
                    npoints_expected,
                    self._progress_log_suffix,
                )
            elif npoints_current:
                self.logger.warning(
                    "only %d/%d points published%s",
                    npoints_current,
                    npoints_expected,
                    self._progress_log_suffix,
                )
            else:
                self.logger.error("no data published%s", self._progress_log_suffix)
        else:
            self.logger.debug(
                "progress %d/%d%s",
                npoints_current,
                npoints_expected,
                self._progress_log_suffix,
            )
        return complete

    def flush(self):
        """Flush any buffered data"""
        self.create()



# -*- coding: utf-8 -*-
#
# This file is part of the bliss project
#
# Copyright (c) 2015-2023 Beamline Control Unit, ESRF
# Distributed under the GNU LGPLv3. See LICENSE for more info.

import math
import itertools
import numpy


def divisorGenerator(n, includeone=True):
    """All divisor of n"""
    for i in range(1, int(math.sqrt(n) + 1)):
        if n % i == 0:
            if includeone or i != 1:
                yield i
            if i * i != n:
                yield n // i


def asproduct(a, n, includeone=True):
    """All tuples of length n with product a"""
    nums = divisorGenerator(a, includeone=includeone)
    for tpl in itertools.product(nums, repeat=n):
        if numpy.prod(tpl) == a:
            yield tpl



import os
import re
import subprocess
import pprint
import difflib
import itertools
from collections import OrderedDict
from typing import Optional, List, Tuple, Any
from concurrent.futures import ThreadPoolExecutor


DIFFER = difflib.Differ()

LINESEPSTR = os.linesep.encode("unicode-escape").decode()


def codefilename(tmpdir) -> str:
    return str(tmpdir.join("test.py"))


def outfilename(tmpdir) -> str:
    return str(tmpdir.join("out.log"))


def errfilename(tmpdir) -> str:
    return str(tmpdir.join("err.log"))


def allfilename(tmpdir) -> str:
    return str(tmpdir.join("all.log"))


def read_file(filename: str) -> str:
    try:
        with open(filename, "r") as f:
            return f.read()
    except FileNotFoundError:
        return ""


def read_files(tmpdir) -> Tuple[str, str, str]:
    return (
        read_file(allfilename(tmpdir)),
        read_file(outfilename(tmpdir)),
        read_file(errfilename(tmpdir)),
    )


def generate_test_script(tmpdir, configlogging: bool = False, **_) -> str:
    lines = ["import sys"]
    if configlogging:
        lines += [
            "import logging",
            "import argparse",
            "from blisswriter.utils import logging_utils",
            "parser = argparse.ArgumentParser()",
            "logging_utils.add_cli_args(parser)",
            "logger = logging.getLogger(__name__)",
            "args = parser.parse_args()",
            "logging_utils.cliconfig(logger, args)",
        ]
    else:
        lines += ["import logging", "logger = logging.getLogger()"]
    lines += [
        'logger.critical("CRITICAL")',
        'logger.error("ERROR")',
        'logger.warning("WARNING")',
        'logger.info("INFO")',
        'logger.debug("DEBUG")',
    ]
    if configlogging:
        lines += [
            'logging_utils.print_out("PRINTOUT")',
            'logging_utils.print_err("PRINTERR")',
        ]
    lines += [
        'print("PRINT")',
        'print("")',
        f'print("A{LINESEPSTR}B{LINESEPSTR}", "C"*10, [1,2,3])',
        f'sys.stdout.write("STDOUTWRITE{LINESEPSTR}")',
        "sys.stdout.flush()",
        f'sys.stderr.write("STDERRWRITE{LINESEPSTR}")',
        "sys.stderr.flush()",
    ]
    filename = codefilename(tmpdir)
    with open(filename, mode="w") as f:
        for line in lines:
            f.write(line + os.linesep)
    return filename


# Expected output of the test script for `expected_std`
EXPECTED_LINES = OrderedDict()
EXPECTED_LINES["CRITICAL"] = "CRITICAL", "FATAL:__main__: "
EXPECTED_LINES["ERROR"] = "ERROR", "ERROR:__main__: "
EXPECTED_LINES["WARNING"] = "WARNING", "WARN :__main__: "
EXPECTED_LINES["INFO"] = "INFO", "INFO :__main__: "
EXPECTED_LINES["DEBUG"] = "DEBUG", "DEBUG:__main__: "
EXPECTED_LINES["PRINTOUT"] = "PRINTOUT", ""
EXPECTED_LINES["PRINTERR"] = "PRINTERR", ""
EXPECTED_LINES["PRINT1"] = "PRINT", ""
EXPECTED_LINES["PRINT2"] = "", ""
EXPECTED_LINES["PRINT3"] = (
    f"A{os.linesep}B{os.linesep} CCCCCCCCCC [1, 2, 3]",
    "",
)
EXPECTED_LINES["STDOUTWRITE"] = "STDOUTWRITE", ""
EXPECTED_LINES["STDERRWRITE"] = "STDERRWRITE", ""

LOGGER_OUT_LINES = ["DEBUG", "INFO"]
LOGGER_ERR_LINES = ["WARNING", "ERROR", "CRITICAL"]  # These have stderr as fallback
LOGGER_LINES = LOGGER_OUT_LINES + LOGGER_ERR_LINES
UTIL_OUT_LINES = ["PRINTOUT"]
UTIL_ERR_LINES = ["PRINTERR"]
UTIL_LINES = UTIL_OUT_LINES + UTIL_ERR_LINES
LINES_NEED_LOGGER = LOGGER_OUT_LINES + UTIL_LINES
STD_OUT_LINES = ["STDOUTWRITE"]
STD_ERR_LINES = ["STDERRWRITE"]
STD_LINES = STD_OUT_LINES + STD_ERR_LINES
PRINT_LINES = ["PRINT1", "PRINT2", "PRINT3"]
OUT_LINES = LOGGER_OUT_LINES + UTIL_OUT_LINES + STD_OUT_LINES + PRINT_LINES
ERR_LINES = LOGGER_ERR_LINES + UTIL_ERR_LINES + STD_ERR_LINES

LOG_LEVELS = LOGGER_LINES


def expected_std(
    outtype: Optional[str] = None,
    file: Optional[bool] = None,
    level: Optional[str] = None,
    std: Optional[bool] = None,
    redirectstd: Optional[bool] = None,
    configlogging: Optional[bool] = None,
    **_,
) -> List[str]:
    """
    :param outtype str: out, err or all
    :param bool file: lines in file or stdout/stderr
    :param level:
    :param bool or tuple std: logging to stdout/stderr enabled
    :param bool or tuple redirectstd: redirect stdout/stderr to dedicated loggers
    """
    lines = []
    if outtype == "out":
        other_stream = ERR_LINES
    elif outtype == "err":
        other_stream = OUT_LINES
    else:
        other_stream = []
        redirectstdout, redirectstderr = redirectstd
    for desc, (msg, prefix) in EXPECTED_LINES.items():
        if desc in other_stream:
            continue
        if configlogging:
            if file:
                # expected log file content
                if other_stream:
                    # out or err log file
                    if not redirectstd:
                        if desc in PRINT_LINES:
                            continue
                        if desc in STD_LINES:
                            continue
                else:
                    # out+err log file
                    if not redirectstdout:
                        if desc in PRINT_LINES:
                            continue
                        if desc in STD_OUT_LINES:
                            continue
                    if not redirectstderr:
                        if desc in STD_ERR_LINES:
                            continue
            else:
                # expected stdout/stderr content
                if not std:
                    if redirectstd:
                        continue
                    if desc in LINES_NEED_LOGGER:
                        continue
            if level_filtered(level, desc):
                # Log level too high
                continue
            msg = prefix + msg
        else:
            # expected stdout/stderr content
            if desc in LINES_NEED_LOGGER:
                continue
            # No logger formatting so no prefix
        lines.append(msg)
    if lines:
        lines.append("")
    return lines


def level_filtered(level: str, desc: str) -> bool:
    levels = LOG_LEVELS
    if desc in levels:
        return desc not in levels[levels.index(level) :]
    return False


def expected_stdout(
    stdout: Optional[bool] = None, redirectstdout: Optional[bool] = None, **kwargs
) -> List[str]:
    return expected_std(outtype="out", std=stdout, redirectstd=redirectstdout, **kwargs)


def expected_stderr(
    stderr: Optional[bool] = None, redirectstderr: Optional[bool] = None, **kwargs
) -> List[str]:
    return expected_std(outtype="err", std=stderr, redirectstd=redirectstderr, **kwargs)


def expected_file(
    logfile: Optional[bool] = None,
    outtype: Optional[str] = None,
    configlogging: Optional[bool] = None,
    stdout: Optional[bool] = None,
    stderr: Optional[bool] = None,
    redirectstdout: Optional[bool] = None,
    redirectstderr: Optional[bool] = None,
    **kwargs,
) -> List[str]:
    if logfile and configlogging:
        if outtype == "out":
            std = stdout
            redirectstd = redirectstdout
        elif outtype == "err":
            std = stderr
            redirectstd = redirectstderr
        else:
            std = stdout, stderr
            redirectstd = redirectstdout, redirectstderr
        return expected_std(
            outtype=outtype,
            configlogging=configlogging,
            file=True,
            std=std,
            redirectstd=redirectstd,
            **kwargs,
        )
    else:
        return []


def expected_fileout(fileout: Optional[bool] = None, **kwargs) -> List[str]:
    return expected_file(outtype="out", logfile=fileout, **kwargs)


def expected_fileerr(fileerr: Optional[bool] = None, **kwargs) -> List[str]:
    return expected_file(outtype="err", logfile=fileerr, **kwargs)


def validate_output(tmpdir, output: str, outtype: str, **kwargs):
    if outtype == "stdout":
        lines = expected_stdout(**kwargs)
    elif outtype == "stderr":
        lines = expected_stderr(**kwargs)
    elif outtype == "fileout":
        lines = expected_fileout(**kwargs)
    elif outtype == "fileerr":
        lines = expected_fileerr(**kwargs)
    elif outtype == "fileall":
        return  # TODO
    lines = os.linesep.join(lines)
    lines = lines.split(os.linesep)
    output = output.split(os.linesep)
    timestamp = r" \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2},\d{3} "
    output = [re.sub(timestamp, ":", s) for s in output]

    if lines != output:
        cmd = " ".join(cliargs(tmpdir, **kwargs))
        errmsg = f"Unexpected {repr(outtype)}"
        errmsg += f"\n Command: {repr(cmd)}"
        errmsg += f"\n\n Options: {pprint.pformat(kwargs, indent=2)}"
        errmsg += "\n\n Difference (-: missing, +: unexpected)"
        errmsg += "\n " + "\n ".join(DIFFER.compare(lines, output))
        raise RuntimeError(errmsg)


def cliargs(
    tmpdir,
    level: Optional[str] = None,
    fileall: Optional[bool] = None,
    fileout: Optional[bool] = None,
    fileerr: Optional[bool] = None,
    stdout: Optional[bool] = None,
    stderr: Optional[bool] = None,
    redirectstdout: Optional[bool] = None,
    redirectstderr: Optional[bool] = None,
    **_,
):
    filename = codefilename(tmpdir)
    args = ["python", filename, "--log=" + level]
    if fileall:
        args.append(f"--logfile={allfilename(tmpdir)}")
    if fileout:
        args.append(f"--logfileout={outfilename(tmpdir)}")
    if fileerr:
        args.append(f"--logfileerr={errfilename(tmpdir)}")
    if not stdout:
        args.append("--nologstdout")
    if not stderr:
        args.append("--nologstderr")
    if redirectstdout:
        args.append("--redirectstdout")
    if redirectstderr:
        args.append("--redirectstderr")
    return args


def generate_output(tmpdir, **kwargs) -> dict:
    lst = cliargs(tmpdir, **kwargs)
    p = subprocess.Popen(lst, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    fileall, fileout, fileerr = read_files(tmpdir)
    return {
        "stdout": stdout.decode(),
        "stderr": stderr.decode(),
        "fileall": fileall,
        "fileout": fileout,
        "fileerr": fileerr,
    }


def run_single_test(args: Tuple[List[str], Tuple[Any], Any]) -> None:
    parameters, values, tmpdir = args
    kwargs = dict(zip(parameters, values))
    generate_test_script(tmpdir, **kwargs)
    result = generate_output(tmpdir, **kwargs)
    for outtype, output in result.items():
        validate_output(tmpdir, output, outtype, **kwargs)


def run_test(tmpdir, configlogging: bool = True) -> None:
    choices = {
        "configlogging": (configlogging,),
        "level": LOG_LEVELS,
        "fileout": (False, True),
        "fileerr": (False, True),
        "fileall": (False, True),
        "stdout": (False, True),
        "stderr": (True,),
        "redirectstdout": (False, True),
        "redirectstderr": (False, True),
    }
    parameters = list(choices.keys())

    def args_generator():
        for i, values in enumerate(itertools.product(*choices.values())):
            tmpdiri = tmpdir.join(str(i))
            os.makedirs(str(tmpdiri))
            yield parameters, values, tmpdiri

    with ThreadPoolExecutor(max_workers=100) as pool:
        list(pool.map(run_single_test, args_generator()))


def test_systemlogging(tmpdir):
    run_test(tmpdir, configlogging=False)


def test_logging(tmpdir):
    run_test(tmpdir, configlogging=True)



"""ESRF data policy unofficial (but still valid)
"""

import os


def masterfile_templates(data_policy=None, dataset_template=None):
    """
    Templates for HDF5 file names relative to the dataset directory

    :returns dict(str):
    """
    templates = dict()
    if dataset_template:
        templates["dataset"] = os.path.extsep.join((dataset_template, "h5"))
    if data_policy == "ESRF":
        templates["dataset_collection"] = os.path.join(
            "..", os.path.extsep.join(("{proposal_name}_{collection_name}", "h5"))
        )
        templates["proposal"] = os.path.join(
            "..", "..", os.path.extsep.join(("{proposal_name}_{beamline}", "h5"))
        )
    return templates



import importlib

_originals = {}


def patch_item(module, attr, newitem):
    try:
        olditem = getattr(module, attr)
    except KeyError:
        pass
    else:
        _originals.setdefault(module.__name__, {}).setdefault(attr, olditem)
    setattr(module, attr, newitem)


def unpatch_item(module, attr):
    try:
        olditem = _originals[module.__name__].get(attr)
    except KeyError:
        pass
    else:
        setattr(module, attr, olditem)


def original(module, attr):
    try:
        return _originals[module.__name__][attr]
    except KeyError:
        return getattr(module, attr)


def patched(module, attr):
    return getattr(module, attr) != original(module, attr)


def patch(name, **kwargs):
    module = importlib.import_module(__package__ + ".patch_" + name)
    module.patch(**kwargs)


def unpatch(name, **kwargs):
    module = importlib.import_module(__package__ + ".patch_" + name)
    module.unpatch(**kwargs)



"""Utilities for raw external data wrapping in HDF5
"""

import os
import numpy
import fabio
from fabio.edfimage import EdfImage
from .io_utils import mkdir
from ..utils.array_order import Order


def add_edf_arguments(filenames, createkwargs=None):
    """
    Arguments for `h5py.create_dataset` to link to EDF data frames.

    :param list(str or tuple) filenames: file names (str) and optionally
                                         image indices (tuple)
    :param dict: result of previous call to append to
    :returns dict:
    :raises RuntimeError: not supported by external datasets
    """
    if not isinstance(createkwargs, dict):
        createkwargs = {}
    stack = isinstance(filenames, list)
    if stack:
        if not filenames:
            return createkwargs
    else:
        filenames = [filenames]
    shape0 = createkwargs.get("frame_shape", tuple())
    for filename in filenames:
        if isinstance(filename, (tuple, list)):
            filename, indices = filename
            if not isinstance(indices, (tuple, list)):
                indices = [indices]
        else:
            indices = []
        if ".edf." in os.path.basename(filename):
            raise RuntimeError(
                "{}: external datasets with compression not supported".format(
                    repr(filename)
                )
            )
        if indices:
            img = fabio.open(filename)
            it = (img._frames[i] for i in indices)
        else:
            it = EdfImage.lazy_iterator(filename)

        for frame in it:
            if frame.swap_needed():
                raise RuntimeError(
                    "{} (frame {}): external datasets do not support byte-swap".format(
                        repr(filename), frame.iFrame
                    )
                )
            compressioni = frame._data_compression
            if compressioni:
                compressioni = compressioni.lower()
            if compressioni == "none":
                compressioni = None
            if compressioni is not None:
                raise RuntimeError(
                    "{} (frame {}): external datasets with compression not supported".format(
                        repr(filename), frame.iFrame
                    )
                )
            shapei = frame.shape
            if len(shapei) == 1:
                # TODO: bug in fabio?
                shapei = 1, shapei[0]
            dtypei = frame.dtype
            start = frame.start
            size = frame.size  # TODO: need compressed size
            external = filename, start, size

            def assertEqual(key, value, evalue):
                if value != evalue:
                    raise RuntimeError(
                        "{} (frame {}): {} = {} instead of {}".format(
                            repr(filename), frame.iFrame, repr(key), value, evalue
                        )
                    )

            if shape0:
                assertEqual("shape", shapei, shape0)
            else:
                createkwargs["frame_shape"] = shape0 = shapei
            if "dtype" in createkwargs:
                assertEqual("dtype", dtypei, createkwargs["dtype"])
            else:
                createkwargs["dtype"] = dtypei
            if "compression" in createkwargs:
                assertEqual("compression", compressioni, createkwargs["compression"])
            else:
                createkwargs["compression"] = compressioni
            if "external" in createkwargs:
                createkwargs["external"].append(external)
            else:
                createkwargs["external"] = [external]
    return createkwargs


def resize(createkwargs, enframes, filename, fillvalue):
    """
    Add/remove external files (before finalization).

    :param dict createkwargs:
    :param enframes shape: number of frames
    :param str filename: in case not enough external files
    :param num fillvalue: in case not enough external files
    :returns int: number of frames skipped
    """
    frame_shape = createkwargs.get("frame_shape", None)
    if not frame_shape:
        raise RuntimeError("The shape of one external frame must be provided")
    nframes = len(createkwargs["external"])
    if nframes > enframes:
        createkwargs["external"] = createkwargs["external"][:enframes]
    elif nframes < enframes:
        if nframes:
            ext = os.path.splitext(createkwargs["external"][0])[-1]
        else:
            ext = ".edf"
        if os.path.splitext(filename)[-1] != ext:
            filename += ext
        if ext == ".edf":
            mkdir(os.path.dirname(filename))
            if not frame_shape:
                frame_shape = 1, 1
            elif len(frame_shape) == 1:
                frame_shape = 1, frame_shape[0]
            EdfImage(data=numpy.full(fillvalue, frame_shape)).write(filename)
        else:
            raise RuntimeError(
                "Dummy file with extension {} not supported".format(repr(ext))
            )
        createkwargs["external"] += [filename] * (enframes - nframes)
    return nframes - enframes


def finalize(createkwargs, addorder=None, fillorder=None, shape=None):
    """
    Finalize external dataset arguments: define shape

    :param dict createkwargs:
    :param tuple shape: scan shape (default: (nframes,))
    :param Order addorder: add order to external
    :param Order fillorder: fill order of shape
    :raises RuntimeError: scan shape does not match number of frames
    """
    nframes = len(createkwargs["external"])
    frame_shape = createkwargs.pop("frame_shape", None)
    if not frame_shape:
        raise RuntimeError("The shape of one external frame must be provided")
    if shape:
        createkwargs["shape"] = shape + frame_shape
        if not isinstance(addorder, Order):
            addorder = Order(addorder)
        if not isinstance(fillorder, Order):
            fillorder = Order(fillorder)
        if fillorder.forder:
            raise ValueError("External HDF5 datasets are always saved in C-order")
        createkwargs["external"] = addorder.swap_list(
            createkwargs["external"], shape, fillorder
        )
    else:
        createkwargs["shape"] = (nframes,) + frame_shape


def add_arguments(filenames, createkwargs=None):
    """
    Arguments for `h5py.create_dataset` to link to data frames.
    The resulting shape will be `shape + frame_shape`

    :param list(str or tuple) filenames: file names (str) and optionally image indices (tuple)
    :param dict createkwargs: result of previous call to append to
    :returns dict:
    """
    if not filenames:
        return
    first = filenames[0]
    if isinstance(first, (tuple, list)):
        first = first[0]
    file_format = os.path.splitext(first)[-1].lower()
    if file_format == ".edf":
        return add_edf_arguments(filenames, createkwargs=createkwargs)
    raise ValueError(f"Unknown external data format {repr(file_format)}")



from contextlib import contextmanager
from typing import Generator


@contextmanager
def capture_exceptions() -> Generator[Generator[None, None, None], None, None]:
    exceptions = list()

    @contextmanager
    def capture() -> Generator[None, None, None]:
        try:
            yield
        except Exception as e:
            exceptions.append(e)

    capture.failed = exceptions
    with capture():
        yield capture

    if exceptions:
        raise exceptions[0]



# -*- coding: utf-8 -*-
#
# This file is part of the NexusWriter project
#
# Code is maintained by the ESRF Data Analysis Unit.
#
# Original author: Wout De Nolf
#
# Copyright (c) 2015-2023 ESRF
#
# Distributed under the terms of the LGPL license.
# See LICENSE.txt for more info.

"""
Bliss writer service

"""

# PROTECTED REGION ID(NexusWriter.system_imports) ENABLED START #
# PROTECTED REGION END #    //  NexusWriter.system_imports

# PyTango imports
import tango
from tango import DebugIt
from tango.server import run
from tango.server import Device
from tango.server import attribute, command
from tango.server import device_property
from tango import AttrQuality, DispLevel, DevState
from tango import AttrWriteType, PipeWriteType
import enum

# Additional import
# PROTECTED REGION ID(NexusWriter.additionnal_import) ENABLED START #
import logging
import re
import os
import itertools
from tango import Attribute

from blisswriter.subscribers import session_subscriber
from blisswriter.subscribers import scan_subscriber
from blisswriter.utils import log_levels
from blisswriter.parameters.base import PROFILE_PARAMETERS
from blisswriter import parameters as writer_parameters

AttrQuality, DispLevel, PipeWriteType  # avoid F401 of pogo generated section


def session_tango_state(state):
    states = session_subscriber.NexusSessionSubscriber._STATE
    if state == states.INIT:
        return DevState.INIT
    elif state == states.ON:
        return DevState.ON
    elif state == states.RUNNING:
        return DevState.RUNNING
    elif state == states.OFF:
        return DevState.OFF
    elif state == states.FAULT:
        return DevState.FAULT
    else:
        return DevState.UNKNOWN


def scan_tango_state(state):
    states = scan_subscriber.NexusScanSubscriber._STATE
    if state == states.INIT:
        return DevState.INIT
    elif state == states.ON:
        return DevState.ON
    elif state == states.OFF:
        return DevState.OFF
    elif state == states.FAULT:
        return DevState.FAULT
    else:
        return DevState.UNKNOWN


def align_listofstrings(lst, alignsep=","):
    """Align a list of string on a separator"""
    if not lst:
        return lst
    if alignsep == " ":
        parts = [re.sub(r"\s+", " ", s).strip().split(alignsep) for s in lst]
    else:
        parts = [s.split(alignsep) for s in lst]
    if len(lst) > 1:
        fmtlst = [
            "{{:>{}}}".format(max(map(len, items)))
            for items in itertools.zip_longest(*parts, fillvalue="")
        ]
        parts = [items + [""] * (len(fmtlst) - len(items)) for items in parts]
        fmtlst[-1] = fmtlst[-1].replace(">", "<")
        fmt = " ".join(fmtlst)
        return [s for items in parts for s in fmt.format(*items).split("\n")]
    else:
        return " ".join(parts[0]).split("\n")


def strftime(tm):
    """
    :param DateTime tm:
    :returns str:
    """
    if tm is None:
        return "not finished"
    else:
        return tm.strftime("%Y-%m-%d %H:%M:%S")


read_log_level = {
    logging.NOTSET: 0,
    logging.DEBUG: 1,
    logging.INFO: 2,
    logging.WARNING: 3,
    logging.ERROR: 4,
    logging.CRITICAL: 5,
}

write_log_level = {v: k for k, v in read_log_level.items()}


read_resource_profiling = {p: i for i, p in enumerate(PROFILE_PARAMETERS)}
write_resource_profiling = {i: p for p, i in read_resource_profiling.items()}


# PROTECTED REGION END #    //  NexusWriter.additionnal_import

__all__ = ["NexusWriter", "main"]


class Resource_profiling(enum.IntEnum):
    """Python enumerated type for Resource_profiling attribute."""

    OFF = 0
    CPU30 = 1
    CPU50 = 2
    CPU100 = 3
    WALL30 = 4
    WALL50 = 5
    WALL100 = 6
    MEM30 = 7
    MEM50 = 8
    MEM100 = 9


class Writer_log_level(enum.IntEnum):
    """Python enumerated type for Writer_log_level attribute."""

    NOTSET = 0
    DEBUG = 1
    INFO = 2
    WARN = 3
    ERROR = 4
    FATAL = 5


class Tango_log_level(enum.IntEnum):
    """Python enumerated type for Tango_log_level attribute."""

    NOTSET = 0
    DEBUG = 1
    INFO = 2
    WARN = 3
    ERROR = 4
    FATAL = 5


class NexusWriter(Device):
    """

    **Properties:**

    - Device Property
        session
            - Bliss session name
            - Type:'DevString'
        keepshape
            - Keep shape of multi-dimensional grid scans
            - Type:'DevBoolean'
        multivalue_positioners
            - Group positioners values
            - Type:'DevBoolean'
        enable_external_nonhdf5
            - Enable external non-hdf5 files like edf (ABSOLUTE LINK!)
            - Type:'DevBoolean'
        disable_external_hdf5
            - Disable external hdf5 files (virtual datasets)
            - Type:'DevBoolean'
        copy_non_external
            - Copy data instead of saving the uri when external linking is disabled
            - Type:'DevBoolean'
        noconfig
            - Do not use extra writer information from Redis
            - Type:'DevBoolean'
        stackmca
            - Merged MCA datasets in application definition
            - Type:'DevBoolean'
        required_disk_space
            - Minimum required disk space in MB to allow the writer to start
            - Type:'DevDouble'
        recommended_disk_space
            - Disk space in MB below which warnings will be generated
            - Type:'DevDouble'
    """

    # PROTECTED REGION ID(NexusWriter.class_variable) ENABLED START #

    def getfromself(self, name):
        """Get python attribute"""
        v = getattr(self, name)
        if isinstance(v, Attribute):
            # Tango attribute
            return getattr(self.proxy, name)
        else:
            # Tango property or python attribute
            return v

    @property
    def saveoptions(self):
        """When an option is not a device property or attribute,
        use the hard-coded default.
        """
        saveoptions = writer_parameters.default_saveoptions()
        for name, info in writer_parameters.all_cli_saveoptions().items():
            if "action" in info:
                store_true = info["action"] == "store_true"
                try:
                    v = self.getfromself(name) == store_true
                except AttributeError:
                    continue
            else:
                try:
                    v = self.getfromself(name)
                except AttributeError:
                    continue
            if name == "resource_profiling":
                # isinstance(v, Resource_profiling) is not True ???
                v = write_resource_profiling[v]
            saveoptions[info["dest"]] = v
        return saveoptions

    # PROTECTED REGION END #    //  NexusWriter.class_variable

    # -----------------
    # Device Properties
    # -----------------

    session = device_property(dtype="DevString", mandatory=True)

    keepshape = device_property(dtype="DevBoolean", default_value=False)

    multivalue_positioners = device_property(dtype="DevBoolean", default_value=False)

    enable_external_nonhdf5 = device_property(dtype="DevBoolean", default_value=False)

    disable_external_hdf5 = device_property(dtype="DevBoolean", default_value=False)

    copy_non_external = device_property(dtype="DevBoolean", default_value=False)

    noconfig = device_property(dtype="DevBoolean", default_value=False)

    stackmca = device_property(dtype="DevBoolean", default_value=False)

    required_disk_space = device_property(dtype="DevDouble", default_value=200)

    recommended_disk_space = device_property(dtype="DevDouble", default_value=1024)

    # ----------
    # Attributes
    # ----------

    resource_profiling = attribute(
        dtype=Resource_profiling,
        access=AttrWriteType.READ_WRITE,
    )

    writer_log_level = attribute(
        dtype=Writer_log_level,
        access=AttrWriteType.READ_WRITE,
    )

    tango_log_level = attribute(
        dtype=Tango_log_level,
        access=AttrWriteType.READ_WRITE,
    )

    scan_states = attribute(
        dtype=("DevState",),
        max_dim_x=10000,
    )

    scan_urls = attribute(
        dtype=("DevString",),
        max_dim_x=10000,
    )

    scan_names = attribute(
        dtype=("DevString",),
        max_dim_x=10000,
    )

    scan_start = attribute(
        dtype=("DevString",),
        max_dim_x=10000,
    )

    scan_end = attribute(
        dtype=("DevString",),
        max_dim_x=10000,
    )

    scan_info = attribute(
        dtype=("DevString",),
        max_dim_x=10000,
    )

    scan_duration = attribute(
        dtype=("DevString",),
        max_dim_x=10000,
    )

    scan_progress = attribute(
        dtype=("DevString",),
        max_dim_x=10000,
    )

    scan_states_info = attribute(
        dtype=("DevString",),
        max_dim_x=10000,
    )

    # ---------------
    # General methods
    # ---------------

    def init_device(self):
        """Initializes the attributes and properties of the NexusWriter."""
        Device.init_device(self)
        self._resource_profiling = Resource_profiling.OFF
        self._writer_log_level = Writer_log_level.NOTSET
        self._tango_log_level = Tango_log_level.NOTSET
        self._scan_states = (DevState.UNKNOWN,)
        self._scan_urls = ("",)
        self._scan_names = ("",)
        self._scan_start = ("",)
        self._scan_end = ("",)
        self._scan_info = ("",)
        self._scan_duration = ("",)
        self._scan_progress = ("",)
        self._scan_states_info = ("",)
        # PROTECTED REGION ID(NexusWriter.init_device) ENABLED START #
        self.proxy = tango.DeviceProxy(self.get_name())
        self.session_writer = getattr(self, "session_writer", None)
        if self.session_writer is None:
            log_levels.init_tango_log_level(device=self)
            self.session_writer = session_subscriber.NexusSessionSubscriber(
                self.session,
                parent_logger=None,
                writer_name="nexus",
                **self.saveoptions,
            )
        self.start()
        # PROTECTED REGION END #    //  NexusWriter.init_device

    def always_executed_hook(self):
        """Method always executed before any TANGO command is executed."""
        # PROTECTED REGION ID(NexusWriter.always_executed_hook) ENABLED START #
        # PROTECTED REGION END #    //  NexusWriter.always_executed_hook

    def delete_device(self):
        """Hook to delete resources allocated in init_device.

        This method allows for any memory or other resources allocated in the
        init_device method to be released.  This method is called by the device
        destructor and by the device Init command.
        """
        # PROTECTED REGION ID(NexusWriter.delete_device) ENABLED START #
        self.session_writer.stop(timeout=3)
        # PROTECTED REGION END #    //  NexusWriter.delete_device

    # ------------------
    # Attributes methods
    # ------------------

    def read_resource_profiling(self):
        # PROTECTED REGION ID(NexusWriter.resource_profiling_read) ENABLED START #
        """Return the resource_profiling attribute."""
        return read_resource_profiling[self.session_writer.resource_profiling]
        # PROTECTED REGION END #    //  NexusWriter.resource_profiling_read

    def write_resource_profiling(self, value):
        # PROTECTED REGION ID(NexusWriter.resource_profiling_write) ENABLED START #
        """Set the resource_profiling attribute."""
        self.session_writer.resource_profiling = write_resource_profiling[value]
        # PROTECTED REGION END #    //  NexusWriter.resource_profiling_write

    def read_writer_log_level(self):
        # PROTECTED REGION ID(NexusWriter.writer_log_level_read) ENABLED START #
        """Return the writer_log_level attribute."""
        return read_log_level[logging.getLogger("blisswriter").getEffectiveLevel()]
        # PROTECTED REGION END #    //  NexusWriter.writer_log_level_read

    def write_writer_log_level(self, value):
        # PROTECTED REGION ID(NexusWriter.writer_log_level_write) ENABLED START #
        """Set the writer_log_level attribute."""
        logging.getLogger("blisswriter").setLevel(write_log_level[value])
        # PROTECTED REGION END #    //  NexusWriter.writer_log_level_write

    def read_tango_log_level(self):
        # PROTECTED REGION ID(NexusWriter.tango_log_level_read) ENABLED START #
        """Return the tango_log_level attribute."""
        return read_log_level[log_levels.tango_get_log_level()]
        # PROTECTED REGION END #    //  NexusWriter.tango_log_level_read

    def write_tango_log_level(self, value):
        # PROTECTED REGION ID(NexusWriter.tango_log_level_write) ENABLED START #
        """Set the tango_log_level attribute."""
        log_levels.tango_set_log_level(write_log_level[value], device=self)
        # PROTECTED REGION END #    //  NexusWriter.tango_log_level_write

    def read_scan_states(self):
        # PROTECTED REGION ID(NexusWriter.scan_states_read) ENABLED START #
        """Return the scan_states attribute."""
        statedict = self.session_writer.scan_state()
        return list(map(scan_tango_state, statedict.values()))
        # PROTECTED REGION END #    //  NexusWriter.scan_states_read

    def read_scan_urls(self):
        # PROTECTED REGION ID(NexusWriter.scan_urls_read) ENABLED START #
        """Return the scan_urls attribute."""
        uscan = self.session_writer.scan_urls()
        lst = [f"{scan}:,{value}" for scan, value in uscan.items()]
        return align_listofstrings(lst)
        # PROTECTED REGION END #    //  NexusWriter.scan_urls_read

    def read_scan_names(self):
        # PROTECTED REGION ID(NexusWriter.scan_names_read) ENABLED START #
        """Return the scan_names attribute."""
        return list(self.session_writer.scan_names())
        # PROTECTED REGION END #    //  NexusWriter.scan_names_read

    def read_scan_start(self):
        # PROTECTED REGION ID(NexusWriter.scan_start_read) ENABLED START #
        """Return the scan_start attribute."""
        data = self.session_writer.scan_start()
        lst = [f"{scan}:,{strftime(value)}" for scan, value in data.items()]
        return align_listofstrings(lst)
        # PROTECTED REGION END #    //  NexusWriter.scan_start_read

    def read_scan_end(self):
        # PROTECTED REGION ID(NexusWriter.scan_end_read) ENABLED START #
        """Return the scan_end attribute."""
        data = self.session_writer.scan_end()
        lst = [f"{scan}:,{strftime(value)}" for scan, value in data.items()]
        return align_listofstrings(lst)
        # PROTECTED REGION END #    //  NexusWriter.scan_end_read

    def read_scan_info(self):
        # PROTECTED REGION ID(NexusWriter.scan_info_read) ENABLED START #
        """Return the scan_info attribute."""
        pscan = self.session_writer.scan_progress_info()
        lst = [
            f"{scan}:{subscan}:,{value}" if len(psubscan) > 1 else f"{scan}:,{value}"
            for scan, psubscan in pscan.items()
            for subscan, value in psubscan.items()
        ]
        return align_listofstrings(lst)
        # PROTECTED REGION END #    //  NexusWriter.scan_info_read

    def read_scan_duration(self):
        # PROTECTED REGION ID(NexusWriter.scan_duration_read) ENABLED START #
        """Return the scan_duration attribute."""
        dscan = self.session_writer.scan_duration()
        lst = [f"{scan}:,{value}" for scan, value in dscan.items()]
        return align_listofstrings(lst)
        # PROTECTED REGION END #    //  NexusWriter.scan_duration_read

    def read_scan_progress(self):
        # PROTECTED REGION ID(NexusWriter.scan_progress_read) ENABLED START #
        """Return the scan_progress attribute."""
        pscan = self.session_writer.scan_progress()
        lst = [
            f"{scan}:{subscan}:,{value}" if len(psubscan) > 1 else f"{scan}:,{value}"
            for scan, psubscan in pscan.items()
            for subscan, value in psubscan.items()
        ]
        return align_listofstrings(lst)
        # PROTECTED REGION END #    //  NexusWriter.scan_progress_read

    def read_scan_states_info(self):
        # PROTECTED REGION ID(NexusWriter.scan_states_info_read) ENABLED START #
        """Return the scan_states_info attribute."""
        data = self.session_writer.scan_state_info()
        lst = [
            f"{name}:,{scan_tango_state(state).name},{reason}"
            for name, (state, reason) in data.items()
        ]
        return align_listofstrings(lst)
        # PROTECTED REGION END #    //  NexusWriter.scan_states_info_read

    # --------
    # Commands
    # --------

    @DebugIt()
    def dev_state(self):
        # PROTECTED REGION ID(NexusWriter.State) ENABLED START #
        """
        This command gets the device state (stored in its device_state data member) and returns it to the caller.

        :return:'DevState'
        Device state
        """
        return session_tango_state(self.session_writer.state)
        # PROTECTED REGION END #    //  NexusWriter.State

    @DebugIt()
    def dev_status(self):
        # PROTECTED REGION ID(NexusWriter.Status) ENABLED START #
        """
        This command gets the device status (stored in its device_status data member) and returns it to the caller.

        :return:'ConstDevString'
        Device status
        """
        return self.session_writer.state_reason
        # PROTECTED REGION END #    //  NexusWriter.Status

    @command(
        dtype_in="DevString",
        doc_in="scan",
        dtype_out="DevState",
    )
    @DebugIt()
    def scan_state(self, argin):
        # PROTECTED REGION ID(NexusWriter.scan_state) ENABLED START #
        """

        :param argin: 'DevString'
        scan_name

        :return:'DevState'
        """
        statedict = self.session_writer.scan_state(argin)
        return scan_tango_state(statedict.get(argin, None))
        # PROTECTED REGION END #    //  NexusWriter.scan_state

    @command(
        dtype_in="DevString",
        doc_in="scan  name",
        dtype_out="DevVarStringArray",
    )
    @DebugIt()
    def scan_url(self, argin):
        # PROTECTED REGION ID(NexusWriter.scan_url) ENABLED START #
        """

        :param argin: 'DevString'
        scan

        :return:'DevVarStringArray'
        """
        uridict = self.session_writer.scan_urls(name=argin)
        return uridict.get(argin, [])
        # PROTECTED REGION END #    //  NexusWriter.scan_url

    @command(
        dtype_in="DevString",
        doc_in="scan name",
        dtype_out="DevBoolean",
    )
    @DebugIt()
    def scan_has_write_permissions(self, argin):
        # PROTECTED REGION ID(NexusWriter.scan_has_write_permissions) ENABLED START #
        """

        :param argin: 'DevString'

        :return:'DevBoolean'
        """
        return self.session_writer.scan_has_write_permissions(argin)
        # PROTECTED REGION END #    //  NexusWriter.scan_has_write_permissions

    @command(
        dtype_in="DevString",
        doc_in="scan name",
        dtype_out="DevBoolean",
    )
    @DebugIt()
    def scan_exists(self, argin):
        # PROTECTED REGION ID(NexusWriter.scan_exists) ENABLED START #
        """

        :param argin: 'DevString'

        :return:'DevBoolean'
        """
        return self.session_writer.scan_exists(argin)
        # PROTECTED REGION END #    //  NexusWriter.scan_exists

    @command(
        dtype_in="DevString",
        doc_in="scan name",
    )
    @DebugIt()
    def stop_scan(self, argin):
        # PROTECTED REGION ID(NexusWriter.stop_scan) ENABLED START #
        """
        Stop a scan writer gracefully (stops processing events and finalizes file writing).

        :param argin: 'DevString'
        scan name

        :return:None
        """
        self.session_writer.stop_scan_writer(argin)
        # PROTECTED REGION END #    //  NexusWriter.stop_scan

    @command()
    @DebugIt()
    def stop(self):
        # PROTECTED REGION ID(NexusWriter.stop) ENABLED START #
        """
        Stop accepting Redis events (start scan, stop scan). Running scan writers will never be stopped automatically (use ``stop_scan(scan_name)``).

        :return:None
        """
        self.session_writer.stop()
        # PROTECTED REGION END #    //  NexusWriter.stop

    @command()
    @DebugIt()
    def start(self):
        # PROTECTED REGION ID(NexusWriter.start) ENABLED START #
        """
        Start accepting Redis events (start scan, stop scan).

        :return:None
        """
        # Fill with device properties (attributes are already set)
        self.session_writer.update_saveoptions(**self.saveoptions)
        self.session_writer.start()
        # PROTECTED REGION END #    //  NexusWriter.start

    @command(
        dtype_in="DevString",
        doc_in="scan name",
        dtype_out="DevString",
    )
    @DebugIt()
    def scan_state_reason(self, argin):
        # PROTECTED REGION ID(NexusWriter.scan_state_reason) ENABLED START #
        """
        The reason why this scan is in this state.

        :param argin: 'DevString'
        scan name

        :return:'DevString'
        """
        statedict = self.session_writer.scan_state_info(name=argin)
        return statedict.get(argin, (None, ""))[1]
        # PROTECTED REGION END #    //  NexusWriter.scan_state_reason

    @command(
        dtype_in="DevString",
        doc_in="absolute path",
    )
    @DebugIt()
    def makedirs(self, argin):
        # PROTECTED REGION ID(NexusWriter.makedirs) ENABLED START #
        """
        Create directories recursively and do not complain when they exist already

        :param argin: 'DevString'

        :return:None
        """
        if not os.path.isabs(argin):
            raise ValueError("Path must be absolute")
        os.makedirs(argin, exist_ok=True)
        # PROTECTED REGION END #    //  NexusWriter.makedirs

    @command()
    @DebugIt()
    def purge_scans(self):
        # PROTECTED REGION ID(NexusWriter.purge_scans) ENABLED START #
        """
        Purge successfully finished scans

        :return:None
        """
        self.session_writer.purge_scan_writers(delay=False)
        # PROTECTED REGION END #    //  NexusWriter.purge_scans

    @command(
        dtype_in="DevString",
        doc_in="scan name",
        dtype_out="DevBoolean",
    )
    @DebugIt()
    def scan_has_required_disk_space(self, argin):
        # PROTECTED REGION ID(NexusWriter.scan_has_required_disk_space) ENABLED START #
        """
        Check whether there is enough disk space for the scan

        :param argin: 'DevString'
        scan name

        :return:'DevBoolean'
        """
        return self.session_writer.scan_has_required_disk_space(argin)
        # PROTECTED REGION END #    //  NexusWriter.scan_has_required_disk_space

    @command(
        dtype_in="DevString",
        doc_in="scan name",
        dtype_out="DevString",
        doc_out="error message",
    )
    @DebugIt()
    def scan_disk_space_error(self, argin):
        # PROTECTED REGION ID(NexusWriter.scan_disk_space_error) ENABLED START #
        """
        Returns a non-empty string when not enough disk space for a scan

        :param argin: 'DevString'
        scan name

        :return:'DevString'
        error message
        """
        return self.session_writer.scan_disk_space_error(argin)
        # PROTECTED REGION END #    //  NexusWriter.scan_disk_space_error

    @command(
        dtype_in="DevString",
        doc_in="scan name",
        dtype_out="DevString",
    )
    @DebugIt()
    def scan_disk_space_warning(self, argin):
        # PROTECTED REGION ID(NexusWriter.scan_disk_space_warning) ENABLED START #
        """
        Returns a non-empty string when not enough disk space for a scan

        :param argin: 'DevString'
        scan name

        :return:'DevString'
        """
        return self.session_writer.scan_disk_space_warning(argin)
        # PROTECTED REGION END #    //  NexusWriter.scan_disk_space_warning


# ----------
# Run server
# ----------

# PROTECTED REGION ID(NexusWriter.custom_code) ENABLED START #
# PROTECTED REGION END #    //  NexusWriter.custom_code


def main(args=None, **kwargs):
    """Main function of the NexusWriter module."""
    # PROTECTED REGION ID(NexusWriter.main) ENABLED START #
    return run((NexusWriter,), args=args, **kwargs)
    # PROTECTED REGION END #    //  NexusWriter.main


# PROTECTED REGION ID(NexusWriter.custom_functions) ENABLED START #
# PROTECTED REGION END #    //  NexusWriter.custom_functions


if __name__ == "__main__":
    main()



import os
import sys
import logging

from tango import DeviceProxy, DevFailed, Database, DbDevInfo

from ..utils import logging_utils
from .utils import config_root_logger

logger = logging.getLogger(__name__)


WRITER_CLASS = "NexusWriter"


def beamline():
    """
    :returns str:
    """
    name = "id00"
    for k in "BEAMLINENAME", "BEAMLINE":
        name = os.environ.get(k, name)
    return name.lower()


def find_session_writers(session_name, db=None):
    """
    Find all TANGO devices of class NexusWriter listening to a particular BLISS session.

    :param std session_name:
    :param Database db:
    :returns list(str): device names
    """
    if db is None:
        db = Database()
    writers = []
    for obj_name in db.get_device_name("*", WRITER_CLASS):
        prop = db.get_device_property(obj_name, "session")["session"]
        if prop:
            if session_name == prop[0]:
                writers.append(obj_name)
    logger.info(
        "Registered writers for session {}: {}".format(repr(session_name), writers)
    )
    return writers


def find_session_writer(session_name, db=None):
    """
    Find TANGO device of class NexusWriter listening to a particular BLISS session.

    :param std session_name:
    :param Database db:
    :returns str or None: device name
    :raises RuntimeError: more than one writer listening to session
    """
    writers = find_session_writers(session_name, db=db)
    if not writers:
        return None
    if len(writers) > 1:
        raise ValueError(
            "Found more than one writer for session {}: {}".format(
                repr(session_name), writers
            )
        )
    return writers[0]


def get_uri(p):
    return "tango://{}:{}/{}".format(p.get_db_host(), p.get_db_port(), p.dev_name())


def ensure_existence(
    session_name,
    server="nexuswriter",
    instance=None,
    domain=None,
    family="bliss_nxwriter",
    member=None,
    use_existing=True,
):
    """
    Find or register TANGO device of class NexusWriter

    :param str session_name:
    :param str server: device server name
    :param str instance: device server instance (Default: `session_name`)
    :param str domain: location of device
    :param str family: type of device
    :param str member: device name (Default: `session_name`)
    :param bool use_existing: use an existing writer for this session (if any)
    :returns DeviceProxy:
    """
    db = Database()
    if not member:
        member = session_name
    if not domain:
        domain = beamline()
    if not instance:
        instance = session_name
    dev_name = "/".join([domain, family, member])
    if use_existing:
        pdev_name = find_session_writer(session_name, db=db)
        if pdev_name:
            proxy = DeviceProxy(pdev_name)
            msg = "{} already registered".format(repr(get_uri(proxy)))
            if dev_name == pdev_name:
                logger.info(msg)
            else:
                logger.warning(msg)
            return proxy
    return register(session_name, dev_name, server=server, instance=instance)


def register(session_name, dev_name, server="nexuswriter", instance="nexuswriters"):
    """
    Register TANGO device of class NexusWriter

    :param str session_name:
    :param str dev_name: for example id00/bliss_nxwriter/test_session
    :param str instance: device server instance
    :returns DeviceProxy:
    """
    # Register proxy is not already registered
    try:
        proxy = DeviceProxy(dev_name)
        logger.info("{} already registered".format(repr(get_uri(proxy))))
    except DevFailed:
        db = Database()
        dev_info = DbDevInfo()
        dev_info.name = dev_name
        dev_info._class = WRITER_CLASS
        server = "/".join([server, instance])
        dev_info.server = server
        db.add_device(dev_info)
        proxy = DeviceProxy(dev_name)
        logger.info("{} registered".format(repr(get_uri(proxy))))
        proxy.put_property({"session": session_name})
    try:
        session = proxy.get_property("session")["session"][0]
    except (IndexError, KeyError):
        session = ""
    if session != session_name:
        raise RuntimeError(
            "{} is listening to Bliss session {} instead of {}".format(
                repr(get_uri(proxy)), repr(session), repr(session_name)
            )
        )
    return proxy


def main(argv=None):
    import argparse

    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(
        description="Register a Tango device saving Bliss data in HDF5"
    )
    parser.add_argument("session", type=str, help="Bliss session name")
    parser.add_argument(
        "--server",
        type=str,
        default="nexuswriter",
        help="Server name ('nexuswriter' by default)",
    )
    parser.add_argument(
        "--instance",
        type=str,
        default="",
        help="Server instance name (session name by default)",
    )
    parser.add_argument(
        "--domain",
        type=str,
        default="",
        help="Device domain name (checks environment or 'id00' by default)",
    )
    parser.add_argument(
        "--family",
        type=str,
        default="bliss_nxwriter",
        help="Device family name ('bliss_nxwriter' by default)",
    )
    parser.add_argument(
        "--member", type=str, default="", help="Device name (session name by default)"
    )
    parser.add_argument(
        "--ignore_existing",
        action="store_false",
        dest="use_existing",
        help="Ignore existing writer for this session",
    )
    logging_utils.add_cli_args(parser)
    args = parser.parse_args(argv[1:])
    config_root_logger(args)
    ensure_existence(
        args.session,
        server=args.server,
        instance=args.instance,
        domain=args.domain,
        family=args.family,
        member=args.member,
        use_existing=args.use_existing,
    )


if __name__ == "__main__":
    sys.exit(main())



"""HDF5 configuration of optimal data storage (IO speed, compression, ...)
"""

from typing import Optional, Tuple
from numbers import Integral
from collections.abc import Mapping

import numpy
from numpy.typing import DTypeLike

try:
    import hdf5plugin
except ImportError:
    hdf5plugin = None

from ..utils.array_size import ShapeType
from ..utils import array_size

VarShapeType = Tuple[array_size.Integral]  # <=0 marks a variable dimension
VarH5pyShapeType = Tuple[
    Optional[array_size.StrictPositiveIntegral]
]  # None marks a variable dimension

DEFAULT_CHUNK_NBYTES = 1 << 20
DEFAULT_COMPRESSION_LIMIT_NBYTES = 1 << 20
DEFAULT_CHUNK_SPLIT = 4
DEFAULT_COMPRESSION_SCHEME = "gzip-byteshuffle"

# Default data size is 2 MB for a 1D detector
# and 2 KB for a 0D detector:
DEFAULT_SCAN_DIM_SIZE = 512
DEFAULT_DETECTOR_DIM_SIZE = 1024
DEFAULT_DTYPE = numpy.int32


def guess_data_shape(
    scan_shape: VarShapeType,
    detector_shape: VarShapeType,
    max_shape: Optional[VarH5pyShapeType],
) -> ShapeType:
    scan_shape = tuple(n if n > 0 else DEFAULT_SCAN_DIM_SIZE for n in scan_shape)
    detector_shape = tuple(
        n if n > 0 else DEFAULT_SCAN_DIM_SIZE for n in detector_shape
    )
    data_shape = scan_shape + detector_shape
    if max_shape:
        assert len(max_shape) == len(
            data_shape
        ), "HDF5 dataset shape must have the same dimensions as maxshape"
        data_shape = tuple(
            n1 if not n2 else max(n1, n2) for n1, n2 in zip(data_shape, max_shape)
        )
    return data_shape


def guess_chunk_shape(
    data_shape: ShapeType,
    dtype: DTypeLike,
    chunk_split: Optional[Integral] = None,
    chunk_nbytes: Optional[Integral] = None,
) -> Optional[ShapeType]:
    """Try to guess the optimal chunk shape with these constraints:
    * Split any dimension for partial access
    * Below the maximal chunk size (1 MB by default, uncompressed)

    The inner-most dimensions are split in `chunk_split` parts
    until chunk_nbytes is reached. The chunk size in the outer
    dimensions will be 1, unless the data size is to small.
    """
    if chunk_nbytes is None:
        chunk_nbytes = DEFAULT_CHUNK_NBYTES
    if chunk_split is None:
        chunk_split = DEFAULT_CHUNK_SPLIT
    itemsize = array_size.dtype_nbytes(dtype)
    size = array_size.shape_to_size(data_shape)
    nbytes = size * itemsize
    if nbytes <= chunk_nbytes:
        return None

    max_size = chunk_nbytes // itemsize
    current_size = 1
    chunk_shape = []
    for n_i in data_shape[-1::-1]:
        if current_size >= max_size:
            c_i = 1
        else:
            a = int(numpy.ceil(n_i / chunk_split))
            b = int(numpy.ceil(max_size / current_size))
            c_i = min(a, b)
        chunk_shape.append(c_i)
        current_size *= c_i
    chunk_shape = tuple(chunk_shape[::-1])
    if chunk_shape == data_shape:
        return None
    return chunk_shape


def guess_compression(
    data_shape: ShapeType,
    dtype: DTypeLike,
    compression_limit_nbytes: Optional[Integral] = None,
) -> bool:
    """Compression is needed when the total data size exceeds the limits (1 MB by default)."""
    if compression_limit_nbytes is None:
        compression_limit_nbytes = DEFAULT_COMPRESSION_LIMIT_NBYTES
    nbytes = array_size.shape_to_nbytes(data_shape, dtype)
    return nbytes > compression_limit_nbytes


def get_compression_arguments(compression_scheme: Optional[str] = None) -> Mapping:
    if compression_scheme:
        compression_scheme = compression_scheme.lower()
    if compression_scheme is None:
        compression_scheme = DEFAULT_COMPRESSION_SCHEME
    if compression_scheme == "none":
        return dict()
    elif compression_scheme == "gzip":
        return {"compression": "gzip"}
    elif compression_scheme == "byteshuffle":
        return {"shuffle": True}
    elif compression_scheme == "gzip-byteshuffle":
        return {"compression": "gzip", "shuffle": True}
    elif compression_scheme == "bitshuffle":
        if hdf5plugin is None:
            raise RuntimeError(
                "Writer does not support HDF5 'bitshuffle' compression. Install the hdf5plugin library"
            )
        return hdf5plugin.Bitshuffle(nelems=0, lz4=False)
    elif compression_scheme == "lz4-bitshuffle":
        if hdf5plugin is None:
            raise RuntimeError(
                "Writer does not support HDF5 'bitshuffle' compression. Install the hdf5plugin library"
            )
        return hdf5plugin.Bitshuffle(nelems=0, lz4=True)
    else:
        raise ValueError(f"Unknown HDF5 compression '{compression_scheme}'")


def guess_dataset_config(
    scan_shape: VarShapeType,
    detector_shape: VarShapeType,
    dtype: Optional[DTypeLike] = None,
    chunk_split: Optional[Integral] = None,
    chunk_nbytes: Optional[Integral] = None,
    compression_limit_nbytes: Optional[Integral] = None,
    compression_scheme: Optional[str] = None,
    max_shape: Optional[VarH5pyShapeType] = None,
) -> dict:
    """Dataset configuration passed the `h5py.Group.create_dataset` for optimal
    storage (IO speed, compression, ...)
    """
    data_shape = guess_data_shape(
        scan_shape=scan_shape, detector_shape=detector_shape, max_shape=max_shape
    )
    if dtype is None:
        dtype = DEFAULT_DTYPE
    chunk_shape = guess_chunk_shape(
        data_shape=data_shape,
        dtype=dtype,
        chunk_split=chunk_split,
        chunk_nbytes=chunk_nbytes,
    )
    config = {"chunks": chunk_shape}
    compression = guess_compression(
        data_shape=data_shape,
        dtype=dtype,
        compression_limit_nbytes=compression_limit_nbytes,
    )
    if compression:
        config.update(get_compression_arguments(compression_scheme=compression_scheme))
    chunking_required = compression or max_shape is not None
    if chunking_required and chunk_shape is None:
        # Do not let h5py guess the chunk size
        config["chunks"] = data_shape
    return config



from typing import Union, Sequence
import numpy

NumericDataType = Sequence[numpy.ndarray]
"""numpy array or a list of numpy arrays (ragged data like diode samples)
"""

LimaStatusType = Sequence[dict]

ScanReferenceType = Sequence[dict]

ChannelDataType = Union[NumericDataType, LimaStatusType, ScanReferenceType]



import h5py
import traceback
from . import monkey
from ..utils.logging_utils import print_err


OPEN_OBJECTS = {}


def _store(obj):
    oid = obj.id
    if oid not in OPEN_OBJECTS:
        stack = traceback.extract_stack()
        stack.pop()
        stack.pop()
        OPEN_OBJECTS[oid] = stack


def _purge():
    for oid in list(OPEN_OBJECTS.keys()):
        if not oid.valid:
            OPEN_OBJECTS.pop(oid)


def _tracking_file(old_class):
    class trackingFile(old_class):
        def __init__(self, *args, **kwargs):
            super(trackingFile, self).__init__(*args, **kwargs)
            _store(self)
            self._keepname = name = uri_from_id(self.id)
            print_open_objects(" AFTER OPENING FILE " + repr(name))

        def close(self, *args, **kwargs):
            super(trackingFile, self).close(*args, **kwargs)
            _purge()
            print_open_objects(" AFTER CLOSING FILE " + repr(self._keepname))

    return trackingFile


def _tracking_group(old_class):
    class trackingGroup(old_class):
        def __init__(self, *args, **kwargs):
            super(trackingGroup, self).__init__(*args, **kwargs)
            _store(self)
            uri = uri_from_id(self.id)
            print_open_objects(" AFTER OPENING GROUP " + repr(uri))

    return trackingGroup


def _tracking_dataset(old_class):
    class trackingDataset(old_class):
        def __init__(self, *args, **kwargs):
            super(trackingDataset, self).__init__(*args, **kwargs)
            _store(self)
            uri = uri_from_id(self.id)
            print_open_objects(" AFTER OPENING DATASET " + repr(uri))

    return trackingDataset


def uri_from_id(oid):
    uri = h5py.h5i.get_name(oid)
    if uri:
        uri = uri.decode()
    else:
        uri = str(oid)
    filename = h5py.h5f.get_name(oid)
    if filename:
        uri = filename.decode() + "::" + uri
    return uri


def type_from_id(oid):
    if isinstance(oid, h5py.h5f.FileID):
        return "file"
    elif isinstance(oid, h5py.h5g.GroupID):
        return "group"
    elif isinstance(oid, h5py.h5d.DatasetID):
        return "dataset"
    else:
        return "object"


def print_open_objects(msg="", **kwargs):
    print_err("\n### {} OPEN HDF5 OBJECTS{}".format(len(OPEN_OBJECTS), msg), **kwargs)
    sep = ""
    for oid, stack in OPEN_OBJECTS.items():
        if not oid.valid:
            continue
        stack = sep.join(traceback.format_list(stack))
        name = repr(uri_from_id(oid))
        otype = type_from_id(oid)
        print_err("\n Open {} {}\n{}{}".format(otype, name, sep, stack), **kwargs)
    print_err("\n### END OPEN HDF5 OBJECTS\n", **kwargs)


def patch(file=True, group=False, dataset=False):
    if file:
        for mod in [h5py, h5py._hl.files]:
            newitem = _tracking_file(monkey.original(mod, "File"))
            monkey.patch_item(mod, "File", newitem)
    if group:
        for mod in [h5py, h5py._hl.group]:
            newitem = _tracking_group(monkey.original(mod, "Group"))
            monkey.patch_item(mod, "Group", newitem)
    if dataset:
        for mod in [h5py, h5py._hl.dataset]:
            newitem = _tracking_dataset(monkey.original(mod, "Dataset"))
            monkey.patch_item(mod, "Dataset", newitem)


def unpatch():
    monkey.unpatch_item(h5py, "File")
    monkey.unpatch_item(h5py, "Group")
    monkey.unpatch_item(h5py, "Dataset")
    monkey.unpatch_item(h5py._hl.files, "File")
    monkey.unpatch_item(h5py._hl.group, "Group")
    monkey.unpatch_item(h5py._hl.dataset, "Dataset")



import logging
from typing import Tuple, Optional
from collections import namedtuple

import numpy

from ..io import nexus
from ..utils.array_size import shape_to_size
from ..utils.logging_utils import CustomLogger


logger = logging.getLogger(__name__)


NXsignal = namedtuple("NXsignal", ["name", "value", "attrs"])
NXaxis = namedtuple("NXaxis", ["name", "value", "attrs"])
NXplotID = namedtuple("NXplotID", ["name", "scan_shape", "detector_shape"])


def split_shape(shape, detector_ndim):
    scan_ndim = len(shape) - detector_ndim
    scan_shape = shape[:scan_ndim]
    detector_shape = shape[scan_ndim:]
    return scan_shape, detector_shape


class NXPlotWriter:
    def __init__(self, nxplotid: NXplotID, suffix: Optional[str] = None):
        """
        :param nxplotid:
        :param suffix: optional suffix which is added to the NXdata
                       name in case of NXdata splitting
        """
        if suffix:
            self.suffix = "_" + suffix
        else:
            self.suffix = ""
        self._signals = list()
        self._axes = list()
        self.nxplotid = nxplotid

    def add_signal(self, signal: NXsignal):
        self._signals.append(signal)

    @property
    def sortkey(self) -> Tuple[str]:
        """Used to sort NXdata groups (needed in case of name collisions)"""
        return tuple(sorted([s.name for s in self._signals]))

    def add_axes(self, positioners, saveorder):
        """
        :param list(Hdf5DatasetWriter) positioners:
        :param SaveOrder saveorder:
        """
        scan_shape = self.nxplotid.scan_shape
        detector_shape = self.nxplotid.detector_shape
        scan_ndim = len(scan_shape)
        detector_ndim = len(detector_shape)
        ndim = scan_ndim + detector_ndim
        if scan_ndim:
            # Scan axes dict
            axes = {}
            for dproxy in positioners:
                with dproxy.open(create=False) as dset:
                    if dset is None:
                        continue
                    axes[dproxy.master_index] = NXaxis(dproxy.linkname, dset, None)

            # TODO: Scatter plot of non-scalar detector
            #       does not have a visualization so no
            #       axes for now
            flattened = scan_ndim != len(axes)
            if flattened and detector_ndim:
                self._axes = []
                return

            # Sort axes
            #   grid plot: fast axis last
            #   scatter plot: fast axis first
            if saveorder.order == "C" and not flattened:
                # Fast axis last
                keys = reversed(sorted(axes))
            else:
                # Fast axis first
                keys = sorted(axes)

            # Dataset creation arguments
            lst = []
            for i, key in enumerate(keys):
                name, value, attrs = axes[key]
                if scan_ndim > 1:  # Grid plot (e.g. image)
                    # Axis values: shape of scan
                    value = value[()]
                    if value.ndim == 1:
                        value = value.reshape(scan_shape)
                    avgdims = tuple(j for j in range(value.ndim) if i != j)
                    # Average along all but axis dimension
                    value = value.mean(avgdims)
                    # Make linear
                    x = numpy.arange(value.size)
                    mask = numpy.isfinite(value)
                    try:
                        m, b = numpy.polyfit(x[mask], value[mask], 1)
                    except Exception:
                        value = numpy.linspace(
                            numpy.nanmin(value), numpy.nanmax(value), value.size
                        )
                    else:
                        value = m * x + b
                else:  # Scatter plot (coordinates and value)
                    if value.ndim > 1:
                        value = {"data": nexus.getUri(value), "shape": scan_shape}
                lst.append(NXaxis(name, value, attrs))
            axes = lst
        else:
            axes = []

        # Add axes for the data dimensions (which are always at the end)
        if ndim - len(axes) == detector_ndim:
            axes += [
                NXaxis("datadim{}".format(i), numpy.arange(detector_shape[i]), None)
                for i in range(detector_ndim)
            ]
        self._axes = axes

    def save(self, nxentry, plotname):
        """
        Create on NXdata with signals and axes.

        :param h5py.Group nxentry:
        :param str plotname:
        :returns h5py.Group:
        """
        h5group = nexus.nxData(nxentry, plotname)
        nexus.nxDataAddSignals(h5group, self._signals)
        if self._axes:
            nexus.nxDataAddAxes(h5group, self._axes)
        return h5group


class NXdataWriter:
    """Manages the creation of NXdata groups. Adding signals with incompatible shapes will result in multiple NXdata's."""

    def __init__(self, plotname, parent_logger=None):
        """
        :param str plotname: NXdata may become "{name}_{suffix}{i}" depending plot splitting
        :param parent_logger:
        """
        self.plotname = plotname
        self._plots = {}  # NXplotID -> NXPlotWriter
        if parent_logger is None:
            parent_logger = logger
        self.logger = CustomLogger(parent_logger, f"NXdata({plotname})")

    def __bool__(self):
        return bool(self._plots)

    def iter_plots(self):
        """Yield the NXPlotWriter's and their unique names

        :yields (str, NXPlotWriter): plot name and plot
        """
        suffixes = [nxplot.suffix for nxplot in self._plots.values()]
        need_suffix = len(suffixes) > 1
        need_index = len(set(suffixes)) != len(suffixes)
        i = 0
        for nxplot in sorted(self._plots.values(), key=lambda nxplot: nxplot.sortkey):
            plotname = nxplot.nxplotid.name
            if need_suffix:
                plotname += nxplot.suffix
                if need_index:
                    i += 1
                    plotname += str(i)
            yield plotname, nxplot

    def save(self, nxentry, default, plotselect):
        """Save all the NXPlotWriter's as NXdata groups

        :param h5py.Group nxentry:
        :param h5py.Group default: the default NXdata group
        :param str plotselect:
        :returns h5py.Group: the default NXdata group
        """
        for plotname, nxplot in self.iter_plots():
            plot = nxplot.save(nxentry, plotname)
            if default is None or nxplot.nxplotid.name == plotselect:
                default = plot
            self.logger.info("Saved")
        return default

    def add_axes(self, positioners, saveorder):
        """Add axes to all NXPlotWriter's. Call this after all signals
        have been added (adding a singal could add another NXPlotWriter).

        :param list(Hdf5DatasetWriter) positioners:
        :param SaveOrder saveorder:
        """
        for nxplot in self._plots.values():
            nxplot.add_axes(positioners, saveorder)

    def add_signal(self, dproxy, grid):
        """Add a dataset to the appropriate NXPlotWriter

        :param Hdf5DatasetWriter dproxy:
        :param bool grid:
        """
        with dproxy.open() as dset:
            if dset is None:
                return
            linkname = dproxy.linkname
            if not linkname:
                dproxy.logger.warning("cannot be linked too")
                return

            # Determine signal shape
            if dproxy.reshaped == grid:
                if not nexus.HASVIRTUAL:
                    dproxy.logger.error(
                        f"Cannot reshape for plot {repr(self.plotname)} due to missing VDS support"
                    )
                    return
                if grid:
                    shape = dproxy.grid_shape
                else:
                    shape = dproxy.flat_shape
                npoints = shape_to_size(dset.shape)
                enpoints = shape_to_size(shape)
                if npoints != enpoints:
                    dproxy.logger.error(
                        f"Cannot reshape {dset.shape} to {shape} for plot {repr(self.plotname)}"
                    )
                    return
            else:
                shape = dset.shape

            # Arguments for dataset creation
            attrs = {}
            scan_shape, detector_shape = split_shape(shape, dproxy.detector_ndim)
            if shape == dset.shape:
                # Same shape so this will be a link
                value = dset
            else:
                # Different shape so this will be a virtual dataset
                scan_ndim = len(scan_shape)
                detector_ndim = len(detector_shape)
                value = {"data": nexus.getUri(dset), "shape": shape}
                interpretation = nexus.nxDatasetInterpretation(
                    scan_ndim, detector_ndim, scan_ndim
                )
                attrs["interpretation"] = interpretation

        # Add the signal
        signal = NXsignal(name=linkname, value=value, attrs=attrs)
        nxplotid = NXplotID(
            name=self.plotname, scan_shape=scan_shape, detector_shape=detector_shape
        )
        signals = self._plots.get(nxplotid)
        if signals is None:
            signals = self._plots[nxplotid] = NXPlotWriter(
                nxplotid, suffix=dproxy.device_type
            )
        signals.add_signal(signal)
